# -*-coding:utf-8 -*-
#
# Apache License, Version 2.0
#
# Copyright (c) 2023 Robottime(Beijing) Technology Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Led driver.

The led driver is a device that provides a simple interface to control the
ws2812b led strip.

Examples:

    >>> from robodyno.components import LedDriver
    >>> from robodyno.interfaces import CanBus
    >>> can = CanBus()
    >>> strip = LedDriver(can)
    >>> strip.set_color(id_=0, color=(255, 0, 0))
    >>> strip.blink(1, color=(0, 255, 0), period=0.5)
    >>> strip.breathe((2, 3), color=(0, 0, 255), period=0.5)
    >>> strip.marquee((4, 7), color=(255, 255, 0), period=1.0)
    >>> strip.clear()
    >>> can.disconnect()
"""

from typing import Union

from robodyno.interfaces import CanBus
from robodyno.components import CanBusDevice
from robodyno.components.config.model import Model


class LedDriver(CanBusDevice):
    """Led driver.

    Attributes:
        id (int): Device id.
        type (Model): Device type.
        fw_ver (float): Firmware version.
    """

    _CMD_GET_VERSION = 0x01
    _CMD_SET_COLOR = 0x02
    _CMD_CONFIG_CAN = 0x03

    def __init__(self, can: CanBus, id_: int = 0x33):
        """Initializes the led driver.

        Args:
            can (CanBus): Can bus instance.
            id_ (int): Device id.

        Raises:
            ValueError: If the device id is invalid.
        """
        super().__init__(can, id_)
        self.get_version(timeout=0.015)
        if self.type is None or self.type != Model.ROBODYNO_LED_DRIVER:
            raise ValueError(f'Device of id {id_} is not a led driver.')

    def config_can_bus(self, new_id: int = None, bitrate: int = 1000000) -> None:
        """Configures the can bus.

        Args:
            new_id (int): New device id.
            bitrate (int): Can bus bitrate.

        Raises:
            ValueError: If the new id is invalid.
        """
        if new_id is None:
            new_id = self.id
        if not 0x01 <= new_id <= 0x3F:
            raise ValueError('New CAN id must be in the range of 0x01-0x3f.')
        bitrate_id = {
            250000: 0,
            500000: 1,
            1000000: 2,
        }.get(bitrate, 2)
        self._can.send(self.id, self._CMD_CONFIG_CAN, 'HH', new_id, bitrate_id)

    def set_color(self, id_: Union[int, tuple], color: tuple) -> None:
        """Sets the color of the led.

        Args:
            id_ (int | tuple): Led id(s). If it is a tuple, it will set the
                color of multiple leds starting from the first id in the tuple
                to the second id in the tuple.
            color (tuple): Led(s) color.
        """
        if isinstance(id_, int):
            id_ = (id_, id_)
        self._can.send(self.id, self._CMD_SET_COLOR, 'BBBBBBe', *id_, *color, 0, 0)

    def blink(self, id_: Union[int, tuple], color: tuple, period: float) -> None:
        """Lets the led blink.

        Args:
            id_ (int | tuple): Led id(s). If it is a tuple, it will blink
                multiple leds starting from the first id in the tuple to the
                second id in the tuple.
            color (tuple): Led(s) color.
            period (float): Half of the blink period in seconds. The led will
                be on for period seconds and off for period seconds.
        """
        if isinstance(id_, int):
            id_ = (id_, id_)
        self._can.send(self.id, self._CMD_SET_COLOR, 'BBBBBBe', *id_, *color, 1, period)

    def breathe(self, id_: Union[int, tuple], color: tuple, period: float) -> None:
        """Lets the led breathe.

        Args:
            id_ (int | tuple): Led id(s). If it is a tuple, it will breath
                multiple leds starting from the first id in the tuple to the
                second id in the tuple.
            color (tuple): Led(s) color.
            period (float): Half of the breath period in seconds. The led will
                be on for period seconds and off for period seconds.
        """
        if isinstance(id_, int):
            id_ = (id_, id_)
        self._can.send(self.id, self._CMD_SET_COLOR, 'BBBBBBe', *id_, *color, 2, period)

    def marquee(self, id_: tuple, color: tuple, period: float) -> None:
        """Turns on the led one by one and then turns off one by one.

        Args:
            id_ (tuple): Led ids. It will control multiple leds turning on and
                off one by one starting from the first id in the tuple to the
                second id in the tuple.
            color (tuple): Led(s) color.
            period (float): Half of the marquee period in seconds. The led will
                be on for period seconds and off for period seconds.

        Raises:
            ValueError: If the id_ is not a tuple.
        """
        if not isinstance(id_, tuple):
            raise ValueError('id_ must be a tuple.')
        self._can.send(self.id, self._CMD_SET_COLOR, 'BBBBBBe', *id_, *color, 3, period)

    def clear(self) -> None:
        """Turns off all the leds."""
        self._can.send(self.id, self._CMD_SET_COLOR, 'BBBBBBe', 0, 255, 0, 0, 0, 0, 0)



# -*-coding:utf-8 -*-
#
# Apache License, Version 2.0
#
# Copyright (c) 2023 Robottime(Beijing) Technology Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provides a cli for robodyno devices."""

from time import sleep

import click
from rich.console import Console
from rich.table import Table
from rich.columns import Columns
from rich.traceback import install
from rich.prompt import Prompt, Confirm, IntPrompt, FloatPrompt

from robodyno import __version__
from robodyno.interfaces import CanBus
from robodyno.components.can_bus.motor import Motor
from robodyno.components.config.model import Model
from robodyno.tools.can_bus_monitor import CanBusMonitor
from robodyno.tools.motor_info_table import motor_info_table
from robodyno.tools.cli_param_types import ID_LIST, BasedIntPrompt


@click.group()
@click.version_option(__version__)
@click.option('--can-bus', '-c', default='can0', help='CAN bus interface channel.')
@click.option('--baudrate', '-b', default=1000000, help='CAN bus baud rate.')
@click.pass_context
def robodyno(ctx: click.Context, can_bus: str, baudrate: int) -> None:
    """Robodyno command line interface."""
    install(suppress=[click])
    ctx.ensure_object(dict)
    ctx.obj['console'] = Console()
    ctx.obj['can'] = CanBus(baudrate, can_bus)
    ctx.obj['can_bitrate'] = baudrate


@robodyno.command(name='list')
@click.pass_context
def list_devices(ctx: click.Context) -> None:
    """List all devices on the CAN bus."""
    console = ctx.obj['console']

    table = Table()
    table.add_column('ID', style='cyan', justify='right', no_wrap=True)
    table.add_column('Model', style='green', justify='left', no_wrap=True)
    table.add_column('Version', style='yellow', justify='left', no_wrap=True)

    for device_id in range(0x3F):
        try:
            main_ver, sub_ver, type_ = ctx.obj['can'].get(
                device_id, 0x01, 'HHI', timeout=0.015
            )
            table.add_row(
                f'0x{device_id:02X}', Model(type_).name, f'{main_ver}.{sub_ver}'
            )
        except TimeoutError:
            continue
    if table.rows:
        console.print(table)
    else:
        console.print('No device found...')


@robodyno.command()
@click.option('--group', '-g', is_flag=True, help='Group messages by CAN ID.')
@click.pass_context
def monitor(ctx: click.Context, group: bool) -> None:
    """Monitor all devices on the CAN bus."""
    console = ctx.obj['console']
    console.print('Press Ctrl+C to exit.')
    m = CanBusMonitor(ctx.obj['can'], group, console)
    m.monitor()


@robodyno.group()
@click.option(
    '--id', '-i', 'id_set', type=ID_LIST, default=0x10, help='Device ID list.'
)
@click.pass_context
def motor(ctx: click.Context, id_set) -> None:
    """Motor commands."""
    ctx.obj['motors'] = []
    console = ctx.obj['console']
    can = ctx.obj['can']
    for device_id in id_set:
        try:
            m = Motor(can, device_id)
            ctx.obj['motors'].append(m)
        except ValueError:
            console.print(f'Device 0x{device_id:02X} not found.')
            continue


@motor.command()
@click.pass_context
def info(ctx: click.Context) -> None:
    """Get motor info."""
    console = ctx.obj['console']
    motors = ctx.obj['motors']
    tables = []
    for m in motors:
        tables.append(motor_info_table(m))
    console.print(Columns(tables, equal=True))


@motor.command()
@click.pass_context
def enable(ctx: click.Context) -> None:
    """Enable motor."""
    motors = ctx.obj['motors']
    for m in motors:
        m.enable()


@motor.command()
@click.pass_context
def disable(ctx: click.Context) -> None:
    """Disable motor."""
    motors = ctx.obj['motors']
    for m in motors:
        m.disable()


@motor.command(name='init')
@click.option('--pos', '-p', type=float, default=0.0, help='Initial position.')
@click.option('--absolute', '-a', is_flag=True, help='Use absolute position.')
@click.option(
    '--save/--not-save',
    ' /-S',
    default=True,
    help='Save the absolute initial position.',
)
@click.pass_context
def init_pos(ctx: click.Context, pos: float, absolute: bool, save: bool) -> None:
    """Initialize motor."""
    console = ctx.obj['console']
    motors = ctx.obj['motors']
    for m in motors:
        if m.fw_ver < 1:
            console.print(
                f'Absolute position is not supported on '
                f'firmware version {m.fw_ver}.'
            )
            return
        if absolute:
            m.init_abs_pos(pos)
            if save:
                m.save()
        else:
            m.init_pos(pos)


@motor.command(name='pos')
@click.argument('position', type=float)
@click.option(
    '--mode',
    '-m',
    type=click.Choice(['direct', 'filter', 'track']),
    default='filter',
    help='Position control mode. Default is filter mode.',
)
@click.option(
    '--bandwidth', '-b', type=float, default=10.0, help='Bandwidth of the filter.'
)
@click.option(
    '--traj-vel',
    '-v',
    type=float,
    default=5.0,
    help='Trajectory velocity of the position tracking control.',
)
@click.option(
    '--traj-acc',
    '--acc',
    type=float,
    default=10.0,
    help='Trajectory acceleration of the position tracking control.',
)
@click.option(
    '--traj-dec',
    '--dec',
    type=float,
    default=10.0,
    help='Trajectory deceleration of the position tracking control.',
)
@click.option('--vel-ff', type=float, default=0.0, help='Velocity feedforward.')
@click.option('--torque-ff', type=float, default=0.0, help='Torque feedforward.')
@click.option('--absolute', '-a', is_flag=True, help='Use absolute position.')
@click.pass_context
def set_pos(
    ctx: click.Context,
    position: float,
    mode: str,
    bandwidth: float,
    traj_vel: float,
    traj_acc: float,
    traj_dec: float,
    vel_ff: float,
    torque_ff: float,
    absolute: bool,
) -> None:
    """Set motor position."""
    motors = ctx.obj['motors']
    for m in motors:
        if mode == 'direct':
            m.position_mode()
        elif mode == 'filter':
            m.position_filter_mode(bandwidth)
        elif mode == 'track':
            m.position_track_mode(traj_vel, traj_acc, traj_dec)
        m.enable()
        if absolute and m.fw_ver >= 1:
            m.set_abs_pos(position, vel_ff, torque_ff)
        else:
            m.set_pos(position, vel_ff, torque_ff)


@motor.command(name='vel')
@click.argument('velocity', type=float)
@click.option(
    '--mode',
    '-m',
    type=click.Choice(['direct', 'ramp']),
    default='ramp',
    help='Velocity control mode. Default is ramp mode.',
)
@click.option(
    '--ramp',
    '-r',
    type=float,
    default=10.0,
    help='Ramp rate of the velocity ramp control.',
)
@click.option('--torque-ff', type=float, default=0.0, help='Torque feedforward.')
@click.pass_context
def set_vel(
    ctx: click.Context, velocity: float, mode: str, ramp: float, torque_ff: float
) -> None:
    """Set motor velocity."""
    motors = ctx.obj['motors']
    for m in motors:
        if mode == 'direct':
            m.velocity_mode()
        elif mode == 'ramp':
            m.velocity_ramp_mode(ramp)
        m.enable()
        m.set_vel(velocity, torque_ff)


@motor.command(name='torque')
@click.argument('torque', type=float)
@click.pass_context
def set_torque(ctx: click.Context, torque: float) -> None:
    """Set motor torque."""
    motors = ctx.obj['motors']
    for m in motors:
        m.torque_mode()
        m.enable()
        m.set_torque(torque)


@motor.command(name='config')
@click.option('--vel-limit', '-v', is_flag=True, help='Set velocity limit.')
@click.option('--current-limit', '-c', is_flag=True, help='Set current limit.')
@click.option('--pid', '-p', is_flag=True, help='Set PID gains.')
@click.option('--heartbeat-rate', '-h', is_flag=True, help='Set heartbeat rate.')
@click.option('--new-id', '-i', is_flag=True, help='Set new ID.')
@click.option('--can-baudrate', '-b', is_flag=True, help='Set CAN baudrate.')
@click.pass_context
def config(
    ctx: click.Context,
    vel_limit: bool,
    current_limit: bool,
    pid: bool,
    heartbeat_rate: bool,
    new_id: bool,
    can_baudrate: bool,
) -> None:
    """Configure motor parameters."""
    console = ctx.obj['console']
    motors = ctx.obj['motors']
    if len(motors) != 1:
        console.print(
            'Please specify only [yellow bold]one[/yellow bold] motor.'
        )
        return
    m = motors[0]
    if not (
        vel_limit or current_limit or pid or heartbeat_rate or new_id or can_baudrate
    ):  # if no option is specified, set all options except can bus configuartions.
        vel_limit = True
        current_limit = True
        pid = True

    if vel_limit:
        vel_limit_now = m.get_vel_limit(0.1)
        m.set_vel_limit(
            FloatPrompt.ask(
                (
                    f'[green bold]Enter[/] new velocity limit, '
                    f'now: [green bold]{vel_limit_now:.2f} rad/s[/]'
                ),
                console=console,
                default=m.default_vel_limit,
            )
        )
    if current_limit:
        current_limit_now = m.get_current_limit(0.1)
        m.set_current_limit(
            FloatPrompt.ask(
                (
                    f'[green bold]Enter[/] new current limit, '
                    f'now: [green bold]{current_limit_now:.1f} A[/]'
                ),
                console=console,
                default=m.default_current_limit,
            )
        )
    if pid:
        pos_kp, vel_kp, vel_ki = m.get_pid(0.1)
        m.set_pid(
            FloatPrompt.ask(
                (
                    f'[green bold]Enter[/] new position kp, '
                    f'now: [green bold]{pos_kp:.2f}[/]'
                ),
                console=console,
                default=m.default_pos_kp,
            ),
            FloatPrompt.ask(
                (
                    f'[green bold]Enter[/] new velocity kp, '
                    f'now: [green bold]{vel_kp:.2f}[/]'
                ),
                console=console,
                default=m.default_vel_kp,
            ),
            FloatPrompt.ask(
                (
                    f'[green bold]Enter[/] new velocity ki, '
                    f'now: [green bold]{vel_ki:.2f}[/]'
                ),
                console=console,
                default=m.default_vel_ki,
            ),
        )
    heartbeat = 1000
    bitrate = ctx.obj['can_bitrate']
    if heartbeat_rate:
        heartbeat = IntPrompt.ask(
            '[green bold]Enter[/] new heartbeat rate im ms',
            console=console,
            default=1000,
        )
        m.config_can_bus(heartbeat=heartbeat, bitrate=bitrate)
    if can_baudrate:
        console.print(
            '[magenta bold]WARNING[/] This will change the CAN baudrate '
            'of the motor [bold]after a reboot[/bold].'
        )
        bitrate = int(
            Prompt.ask(
                '[green bold]Enter[/] new CAN bitrate',
                console=console,
                choices=['1000000', '500000', '250000'],
                default=str(bitrate),
            )
        )
        m.config_can_bus(
            heartbeat=heartbeat,
            bitrate=bitrate,
        )
    if new_id:
        console.print(
            '[yellow bold]NOTE[/] This will change the device ID of the '
            'motor [bold]immediately[/bold].'
        )
        new_id = BasedIntPrompt.ask(
            '[green bold]Enter[/] new ID',
            console=console,
            default=m.id,
        )
        m.config_can_bus(new_id=new_id, heartbeat=heartbeat, bitrate=bitrate)
        m = Motor(ctx.obj['can'], new_id)
    if Confirm.ask('[cyan bold]Save[/] the new parameters to flash?'):
        m.save()


@motor.command(name='calibrate')
@click.pass_context
def calibrate(ctx: click.Context) -> None:
    """Calibrate motor."""
    console = ctx.obj['console']
    motors = ctx.obj['motors']
    if len(motors) != 1:
        console.print(
            'Please specify only [yellow bold]one[/yellow bold] motor.'
        )
        return
    if not Confirm.ask(
        '[magenta bold]WARNING[/] Calibration will move the motor and '
        'reset the encoders. Make sure the motor is free to move.'
    ):
        return
    m = motors[0]
    with console.status(f'[bold green]Calibrating motor 0x{m.id:02X}...'):
        m.calibrate()
        sleep(1)
        while m.state != Motor.MotorState.DISABLED:
            pass
        if m.error['error'] != Motor.MotorError.NONE:
            console.print(f'[red bold]Error:[/] {m.error}')
            return
        sleep(0.2)
        console.print('[green bold]Done![/]')
    if Confirm.ask('[cyan bold]Save[/] the result to flash?', default=True):
        m.save()


@motor.command(name='clear-errors')
@click.pass_context
def clear_errors(ctx: click.Context) -> None:
    """Clear errors on motors."""
    motors = ctx.obj['motors']
    for m in motors:
        m.clear_errors()


@motor.command(name='reboot')
@click.pass_context
def reboot(ctx: click.Context) -> None:
    """Reboot motors."""
    motors = ctx.obj['motors']
    for m in motors:
        m.reboot()


@motor.command(name='estop')
@click.pass_context
def estop(ctx: click.Context) -> None:
    """Emergency stop motors."""
    motors = ctx.obj['motors']
    for m in motors:
        m.estop()


@motor.command(name='reset')
@click.pass_context
def reset(ctx: click.Context) -> None:
    """Reset motors."""
    console = ctx.obj['console']
    motors = ctx.obj['motors']
    if len(motors) != 1:
        console.print(
            'Please specify only [yellow bold]one[/yellow bold] motor.'
        )
        return
    if not Confirm.ask(
        '[magenta bold]WARNING[/] Resetting the motor will reset '
        '[red bold]ALL[/red bold] parameters to their default values, calibrate '
        'the motor and reboot it. Continue?'
    ):
        return
    m = motors[0]
    with console.status(f'[bold green]Resetting motor 0x{m.id:02X}...'):
        m.reset()
        sleep(4)
        m = Motor(ctx.obj['can'], 0x10)
        console.print(
            f'[green bold]Calibrating[/green bold] motor 0x{m.id:02X}...'
        )
        m.calibrate()
        sleep(1)
        while m.state != Motor.MotorState.DISABLED:
            pass
        if m.error['error'] != Motor.MotorError.NONE:
            console.print(f'[red bold]Error:[/] {m.error}')
            return
        sleep(0.2)
        m.save()
        sleep(0.2)
        console.print(
            f'[green bold]Rebooting[/green bold] motor 0x{m.id:02X}...'
        )
        m.reboot()
        sleep(3)
        console.print('[green bold]Done.[/]')
        console.print(
            'The ID of the motor is now [green bold]0x10[/green bold].'
        )



# -*-coding:utf-8 -*-
#
# Apache License, Version 2.0
#
# Copyright (c) 2023 Robottime(Beijing) Technology Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Stepper develop board driver.

The stepper develop board provides a stepper motor driver with a CAN bus interface.
It can be used to configure the stepper motor driver and control the stepper motor.

Examples:

    >>> from robodyno.components import StepperDriver
    >>> from robodyno.interfaces import CanBus
    >>> can = CanBus()
    >>> stepper = StepperDriver(can)
    >>> stepper.enable()
    >>> stepper.set_vel(1)
    >>> can.disconnect()
"""

from math import pi
from typing import Optional

from robodyno.interfaces import CanBus
from robodyno.components import CanBusDevice
from robodyno.components.config.model import Model


class StepperDriver(CanBusDevice):
    """Stepper develop board driver.

    Attributes:
        id (int): Device id.
        type (Model): Device type.
        fw_ver (float): Firmware version.
    """

    _CMD_SET_NODE_ID = 0x02
    _CMD_SET_SUBDIVISION = 0x03
    _CMD_SET_VEL_ACC_LIMIT = 0x04
    _CMD_ENABLE = 0x05
    _CMD_DISABLE = 0x06
    _CMD_STOP = 0x07
    _CMD_SET_POSITION = 0x08
    _CMD_SET_VELOCITY = 0x09
    _CMD_GET_POSITION = 0x0A
    _CMD_GET_VELOCITY = 0x0B

    def __init__(self, can: CanBus, id_: int = 0x22, reduction: float = 1):
        """Initializes the stepper develop board driver.

        Args:
            can (CanBus): Can bus interface.
            id_ (int): Device id.
            reduction (float): The reduction ratio of the stepper motor.

        Raises:
            ValueError: If the device is not a stepper develop board.
        """
        super().__init__(can, id_)
        self.get_version(timeout=0.015)
        if self.type is None or self.type != Model.ROBODYNO_STEPPER_DRIVER:
            raise ValueError('The device is not a stepper develop board.')
        self._reduction = reduction
        self.set_subdivision(8)

    def config_can_bus(self, new_id: int) -> None:
        """Configures the CAN bus.

        Args:
            new_id (int): New device id.

        Raises:
            ValueError: If the new id is not in the range of 0x01-0x3f.
        """
        if not 0x01 <= new_id <= 0x3F:
            raise ValueError('New CAN id must be in the range of 0x01-0x3f.')
        self._can.send(self.id, self._CMD_SET_NODE_ID, 'B', new_id)

    def set_subdivision(self, subdivision: int) -> None:
        """Sets the subdivision of the stepper motor.

        Args:
            subdivision (int): The subdivision of the stepper motor.

        Raises:
            ValueError: If the subdivision is not 8, 16, 32 or 64.
        """
        if subdivision not in [8, 16, 32, 64]:
            raise ValueError('Subdivision must be 8, 16, 32 or 64.')
        self._subdivision = subdivision
        self._factor = 200.0 * subdivision * self._reduction / 2.0 / pi
        self._can.send(self.id, self._CMD_SET_SUBDIVISION, 'B', subdivision)

    def set_vel_acc_limit(self, max_vel: float, acc: Optional[float] = None) -> None:
        """Sets the maximum velocity and acceleration of the stepper motor.

        Args:
            max_vel (float): The maximum velocity of the stepper motor.
            acc (float): The acceleration of the stepper motor. If None, the
                acceleration is set to 4 times the maximum velocity.
        """
        if acc is None:
            acc = max_vel * 4
        self._can.send(
            self.id,
            self._CMD_SET_VEL_ACC_LIMIT,
            'ff',
            max_vel * self._factor,
            acc * self._factor,
        )

    def enable(self) -> None:
        """Enables the stepper motor."""
        self._can.send(self.id, self._CMD_ENABLE, '')

    def disable(self) -> None:
        """Disables the stepper motor."""
        self._can.send(self.id, self._CMD_DISABLE, '')

    def stop(self) -> None:
        """Stops the stepper motor."""
        self._can.send(self.id, self._CMD_STOP, '')

    def set_pos(self, pos: float) -> None:
        """Sets the position of the stepper motor.

        Args:
            pos (float): The position of the stepper motor in radians.
        """
        self._can.send(self.id, self._CMD_SET_POSITION, 'i', int(pos * self._factor))

    def set_vel(self, vel: float) -> None:
        """Sets the velocity of the stepper motor.

        Args:
            vel (float): The velocity of the stepper motor in rad/s.
        """
        self._can.send(self.id, self._CMD_SET_VELOCITY, 'f', vel * self._factor)

    def get_pos(self, timeout: Optional[float] = None) -> Optional[float]:
        """Reads the position of the stepper motor.

        Args:
            timeout (float): Timeout in seconds.

        Returns:
            (float | None): The position of the stepper motor in radians. Returns
                None if the read times out.
        """
        try:
            pos = self._can.get(self.id, self._CMD_GET_POSITION, 'i', timeout)[0]
        except TimeoutError:
            return None
        return pos / self._factor

    def get_vel(self, timeout: Optional[float] = None) -> Optional[float]:
        """Reads the velocity of the stepper motor.

        Args:
            timeout (float): Timeout in seconds.

        Returns:
            (float | None): The velocity of the stepper motor in rad/s. Returns
                None if the read times out.
        """
        try:
            vel = self._can.get(self.id, self._CMD_GET_VELOCITY, 'f', timeout)[0]
        except TimeoutError:
            return None
        return vel / self._factor



# -*-coding:utf-8 -*-
#
# Apache License, Version 2.0
#
# Copyright (c) 2023 Robottime(Beijing) Technology Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Slider module.

This module provides functions to control slider module in Webots. The slider module 
is a lead screw driven linear module with a robodyno motor.

Examples:

```python
from robodyno.components import SliderModule
from robodyno.interfaces import Webots
webots = Webots()
slider = SliderModule(webots, 0x10)
slider.enable()
slider.set_pos(0.1)
webots.step(10)
print(slider.get_pos())
slider.disable()
webots.stop()
```
"""

from math import fabs, pi
from typing import Optional

from robodyno.interfaces import Webots
from robodyno.components.webots.motor import Motor


class SliderModule(object):
    """Slider module.

    Attributes:
        motor (Motor): Motor object of the slider module.
    """

    def __init__(
        self,
        webots: Webots,
        id_: int = 0x10,
        type_: Optional[str] = None,
        max_vel: float = 0.02,
        lead: float = 0.01,
    ):
        """Initialize the slider module.

        Args:
            webots (Webots): Webots object.
            id_ (int): ID of the slider module.
            type_ (string): Type of the slider module.
            max_vel (float): Maximum velocity of the slider module in m/s.
            lead (float): Lead of the slider module in meter.
        """
        self._lead = lead
        self.motor = Motor(webots, id_, type_)
        self._reduction = self._lead / 2 / pi  # in m/rad
        self.set_max_vel(max_vel)

    def set_max_vel(self, max_vel: float) -> None:
        """Sets max velocity of the slider module.

        This function configures the motor to position track mode with max
        velocity.

        Args:
            max_vel (float): Max velocity of the slider module in m/s.
        """
        self._max_vel = fabs(max_vel)
        self.motor.position_track_mode(self._max_vel / self._reduction, 40, 40)

    def enable(self) -> None:
        """Enables the slider module."""
        self.motor.enable()

    def disable(self) -> None:
        """Disables the slider module."""
        self.motor.disable()

    def set_pos(self, pos: float) -> None:
        """Sets position of the slider module.

        Args:
            pos (float): Position of the slider module in m.
        """
        self.motor.set_pos(pos / self._reduction)

    def set_abs_pos(self, pos: float) -> None:
        """Sets absolute position of the slider module.

        Args:
            pos (float): Absolute position of the slider module in m.
        """
        self.motor.set_abs_pos(pos / self._reduction)

    def get_pos(self, timeout: Optional[float] = None) -> Optional[float]:
        """Reads position of the slider module.

        Timeout is not supported in Webots.

        Returns:
            (float): Position of the slider module in m.
        """
        del timeout
        pos = self.motor.get_pos()
        return pos * self._reduction

    def get_abs_pos(self, timeout: Optional[float] = None) -> Optional[float]:
        """Reads absolute position of the slider module.

        Timeout is not supported in Webots.

        Returns:
            (float): Absolute position of the slider module in m.
        """
        del timeout
        pos = self.motor.get_abs_pos()
        return pos * self._reduction



# -*-coding:utf-8 -*-
#
# Apache License, Version 2.0
#
# Copyright (c) 2023 Robottime(Beijing) Technology Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Slider module.

This module provides functions to control slider module. The slider module is a
lead screw driven linear module with a robodyno motor.

Examples:

    >>> from robodyno.components import SliderModule
    >>> from robodyno.interfaces import CanBus
    >>> can_bus = CanBus()
    >>> slider = SliderModule(can_bus, 0x10)
    >>> slider.enable()
    >>> slider.set_pos(0.1)
    >>> slider.get_pos()
    0.1
    >>> slider.disable()
    >>> can_bus.disconnect()
"""

from math import fabs, pi
from typing import Optional

from robodyno.interfaces import CanBus
from robodyno.components.can_bus.motor import Motor


class SliderModule(object):
    """Slider module.

    Attributes:
        motor (Motor): Motor object of the slider module.
    """

    def __init__(
        self,
        can: CanBus,
        id_: int = 0x10,
        type_: Optional[str] = None,
        max_vel: float = 0.02,
        lead: float = 0.01,
    ):
        """Inits SliderModule with can bus object and device id.

        This function initializes the slider module by creating a motor object
        and configuring the motor to position track mode with max velocity.

        Args:
            can (CanBus): CanBus object.
            id_ (int): Motor id of the slider module.
            type_ (string): Motor type of the slider module.
            max_vel (float): Max velocity of the slider module in m/s.
            lead (float): Lead of the slider module in m/round.
        """
        self._lead = lead
        self.motor = Motor(can, id_, type_)
        self._reduction = self._lead / 2 / pi  # in m/rad
        self.set_max_vel(max_vel)

    def set_max_vel(self, max_vel: float) -> None:
        """Sets max velocity of the slider module.

        This function configures the motor to position track mode with max
        velocity.

        Args:
            max_vel (float): Max velocity of the slider module in m/s.
        """
        self._max_vel = fabs(max_vel)
        self.motor.position_track_mode(self._max_vel / self._reduction, 40, 40)

    def enable(self) -> None:
        """Enables the slider module."""
        self.motor.enable()

    def disable(self) -> None:
        """Disables the slider module."""
        self.motor.disable()

    def init_pos(self, initial_pos: float = 0) -> None:
        """Initializes position of the slider module.

        Args:
            initial_pos (float): Initial position of the slider module in m.
        """
        self.motor.init_pos(initial_pos / self._reduction)

    def init_abs_pos(self, initial_pos: float = 0, save: bool = True) -> None:
        """Initializes absolute position of the slider module.

        Args:
            initial_pos (float): Initial absolute position of the slider module in m.
            save (bool): Whether to save the configuration.
        """
        self.motor.init_abs_pos(initial_pos / self._reduction)
        if save:
            self.motor.save()

    def set_pos(self, pos: float) -> None:
        """Sets position of the slider module.

        Args:
            pos (float): Position of the slider module in m.
        """
        self.motor.set_pos(pos / self._reduction)

    def set_abs_pos(self, pos: float) -> None:
        """Sets absolute position of the slider module.

        Args:
            pos (float): Absolute position of the slider module in m.
        """
        self.motor.set_abs_pos(pos / self._reduction)

    def get_pos(self, timeout: Optional[float] = None) -> Optional[float]:
        """Reads position of the slider module.

        Args:
            timeout (float): Timeout in seconds.

        Returns:
            (float | None): Position of the slider module in m. Returns None if
                the read times out.
        """
        pos = self.motor.get_pos(timeout)
        if pos is None:
            return None
        return pos * self._reduction

    def get_abs_pos(self, timeout: Optional[float] = None) -> Optional[float]:
        """Reads absolute position of the slider module.

        Args:
            timeout (float): Timeout in seconds.

        Returns:
            (float | None): Absolute position of the slider module in m. Returns
                None if the read times out.
        """
        pos = self.motor.get_abs_pos(timeout)
        if pos is None:
            return None
        return pos * self._reduction



import time
from math import fabs

def linear_interpolation(_from, _to, speed = None, duration = 0):
    """Linear Interpolation Algorism
    
    Args:
        _from: start value
        _to: end value
        speed: (/s), motion speed
        duration: (s), motion duration when speed not set
    
    Yield:
        interpolated value
    """
    start_t = time.time()
    dur = duration
    if speed:
        dur = fabs((_to - _from) / speed)
    while dur > 0:
        dt = time.time() - start_t
        if dt > dur:
            break
        yield((_to - _from) * dt / dur + _from)
    yield(_to)


# -*-coding:utf-8 -*-
#
# The MIT License (MIT)
#
# Copyright (c) 2023 Robottime(Beijing) Technology Co., Ltd
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""This module provides a class for controlling Robodyno Six Dof Collaborative Robot.

The SixDoFCollabRobot class provided by this module is used to control Robodyno Six Dof Collaborative Robot
through the CAN bus. It provides methods for setting Six Dof Collaborative Robot parameters, reading Six Dof Collaborative Robot
states, and controlling the Six Dof Collaborative Robot to run in different space.

Examples:

    >>> from robodyno.components import Motor
    >>> from robodyno.interfaces import CanBus
    >>> from robodyno.robots.six_dof_collaborative_robot import SixDoFCollabRobot
    >>> can = CanBus()
    >>> class MySixDoFArm(SixDoFCollabRobot):
    >>>     def __init__(self):
    >>>         M1 = Motor(can, 0x10, 'ROBODYNO_PRO_44')
    >>>         M2 = Motor(can, 0x11, 'ROBODYNO_PRO_44')
    >>>         M3 = Motor(can, 0x12, 'ROBODYNO_PRO_44')
    >>>         M4 = Motor(can, 0x13, 'ROBODYNO_PRO_44')
    >>>         M5 = Motor(can, 0x14, 'ROBODYNO_PRO_12')
    >>>         M6 = Motor(can, 0x15, 'ROBODYNO_PRO_12')
    >>>     super().__init__(M1, M2, M3, M4, M5, M6, 0.18, 0.135, 0.135, 0.075, 0.075, 0.1)
    >>> arm = MySixDoFArm()
    >>> arm.init()
    >>> arm.get_joints_poses()
    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
    >>> can.disconnect()
"""

import time
import numpy as np
from cmath import acos, asin
from math import pi, sqrt, sin, cos, atan2
from ..utils.transformations import dh_matrix, euler_from_matrix, translation_from_matrix, translation_matrix, euler_matrix
from ..utils.interpolations import linear_interpolation

class SixDoFCollabRobot(object):
    """Controls Robodyno Six Dof Collaborative Robot through the CAN bus.
    
    Attributes:
        joints (list): list of 6 joint motors
        l01 (float): link from world to joint 1 (m) 
        l12 (float): link from joint 1 to joint 2 (m) 
        l23 (float): link from joint 2 to joint 3 (m) 
        l34 (float): link from joint 3 to joint 4 (m) 
        l45 (float): link from joint 4 to joint 5 (m) 
        l56 (float): link from joint 5 to joint 6 (m) 
        end_effector (object): end effector object 
    """
    
    def __init__(self, j1, j2, j3, j4, j5, j6, 
                 l01: float, l12: float, l23: float, l34: float, 
                 l45: float, l56: float, end_effector: object = None):
        """Initializes robot with joints and links
        
        Args:
            j1 (Motor): base_motor.
            j2 (Motor): shoulder_motor.
            j3 (Motor): upperarm_motor.
            j4 (Motor): elbow_motor.
            j5 (Motor): forearm_motor.
            j6 (Motor): hand_motor.
            l01 (float): link from world to joint 1 (m)
            l12 (float): link from joint 1 to joint 2 (m)
            l23 (float): link from joint 2 to joint 3 (m)
            l34 (float): link from joint 3 to joint 4 (m)
            l45 (float): link from joint 4 to joint 5 (m)
            l56 (float): link from joint 5 to joint 6 (m)
            end_effector (object): end effector object.
        """
        self.joints = [j1, j2, j3, j4, j5, j6]
        self.l01 = l01
        self.l12 = l12
        self.l23 = l23
        self.l34 = l34
        self.l45 = l45
        self.l56 = l56
        self.end_effector = end_effector

        for m in self.joints:
            m.position_filter_mode(8)
        self._axes_poses = [0 for i in range(6)]
        self._axes_zeros = [0 for i in range(6)]

    def get_joints_poses(self) -> list:
        """Read joints positions to a list.
        
        Returns:
            (list): a list of 6 joints positions
            
        Raises:
            RuntimeError: If the motor Joint is invalid.
        """
        poses = []
        for i in range(6):
            pos = None
            count = 0
            while pos is None:
                if count > 5:
                    raise RuntimeError('Filed to get position of Joint {}.'.format(i+1))
                count += 1
                pos = self.joints[i].get_pos(0.2)
            poses.append(pos - self._axes_zeros[i])
        return poses

    def enable(self) -> None:
        """enable joints motors"""
        for i in range(6):
            self.joints[i].enable()
    
    def disable(self) -> None:
        """disable joints motors"""
        for i in range(6):
            self.joints[i].disable()
    
    def init(self, axes_poses: list = [0 for i in range(6)]) -> None:
        """Calibrate robot motors with given axes poses.
        
        Args:
            axes_poses (list): list of 6 axes poses(rad)
        """
        self._axes_poses = axes_poses.copy()
        self._axes_zeros = [0 for i in range(6)]
        cur_poses = self.get_joints_poses()
        for i in range(6):
            self._axes_zeros[i] = cur_poses[i] - self._axes_poses[i]
            self._axes_poses[i] = 0

    def set_joint_pos(self, id: int, pos: float) -> None:
        """Set joint angle with joint id and position
        
        Args:
            id (int): joint id (0-5)
            pos (float): joint target position(rad)
        """
        self.joints[id].set_pos(pos + self._axes_zeros[id])
        self._axes_poses[id] = pos
    
    def joint_space_interpolated_motion(self, target: list, 
                                        speeds: list = [None for i in range(6)], 
                                        duration:float = 0) -> None:
        """Robot interpolated motion in joint space.
        
        Args:
            target (list): iterable of 6 joints target angle(rad)
            speeds (list): iterable of 6 joints motion speed(rad/s)
            duration (float): default motion duration(s)
        """
        interpolations = []
        joint_poses = self._axes_poses.copy()
        for i in range(6):
            interpolations.append(linear_interpolation(joint_poses[i], target[i], speed = speeds[i], duration=duration))
        update_flag = True
        while update_flag:
            update_flag = False
            for i in range(6):
                pos = next(interpolations[i], None)
                if pos is not None:
                    self.set_joint_pos(i, pos)
                    update_flag = True
            time.sleep(0.05)

    def home(self, duration: float = 5) -> None:
        """Move back to zero position.
        
        Args:
            duration (float): motion duration(s)
        """
        self.joint_space_interpolated_motion((0,0,0,0,0,0), duration = duration)

    def cartesian_space_interpolated_motion(self, x: float, y: float, z, roll: float, pitch: float, yaw: float, 
                                            sol_id: int = 2, x_speed: float = None, y_speed: float = None, z_speed: float = None, 
                                            roll_speed: float = None, pitch_speed: float = None, yaw_speed: float = None, duration: float = 0) -> None:
        """Robot Interpolated motion in Cartesian space.
        
        Args:
            x (float): target robot end x
            y (float): target robot end y
            z (float): target robot end z
            roll (float): target robot end roll
            pitch (float): target robot end pitch
            yaw (float): target robot end yaw
            sol_id (int): inverse kinematics solve id (0-7)
            x_speed (float): speed alone X dimension(m/s)
            y_speed (float): speed alone Y dimension(m/s)
            z_speed (float): speed alone Z dimension(m/s)
            roll_speed (float): rotation speed on X axis(rad/s)
            pitch_speed (float): rotation speed on Y axis(rad/s)
            yaw_speed (float): rotation speed on Z axis(rad/s)
            duration (float): default motion duration(s)
        """
        current_x, current_y, current_z, current_roll, current_pitch, current_yaw =  self.forward_kinematics(self._axes_poses.copy())

        x_interpolation = linear_interpolation(current_x, x, x_speed, duration)
        y_interpolation = linear_interpolation(current_y, y, y_speed, duration)
        z_interpolation = linear_interpolation(current_z, z, z_speed, duration)
        roll_interpolation = linear_interpolation(current_roll, roll, roll_speed, duration)
        pitch_interpolation = linear_interpolation(current_pitch, pitch, pitch_speed, duration)
        yaw_interpolation = linear_interpolation(current_yaw, yaw, yaw_speed, duration)
        while True:
            temp_x = next(x_interpolation, x)
            temp_y = next(y_interpolation, y)
            temp_z = next(z_interpolation, z)
            temp_roll = next(roll_interpolation, roll)
            temp_pitch = next(pitch_interpolation, pitch)
            temp_yaw = next(yaw_interpolation, yaw)
            axes = self.inverse_kinematics(temp_x, temp_y, temp_z, temp_roll, temp_pitch, temp_yaw, sol_id)
            for i in range(6):
                self.set_joint_pos(i, axes[i])
            if temp_x == x and temp_y == y and temp_z == z and temp_roll == roll and temp_pitch == pitch and temp_yaw == yaw:
                break
            time.sleep(0.05)

    def forward_kinematics(self, axes: list) -> tuple:
        """Forward kinematics algorism
        
        Args:
            axes (list): list of 6 joint angles(rad)
        
        Returns:
            (tuple): (x, y, z, roll, pitch, yaw) tuples of robot end transform
            
        """
        T01 = dh_matrix(pi/2 , 0 , self.l01, axes[0])
        T12 = dh_matrix(0    , -self.l12, 0 , axes[1]-pi/2)
        T23 = dh_matrix(0    , -self.l23, 0 , -axes[2])
        T34 = dh_matrix(pi/2 , 0 , self.l34, axes[3]-pi/2)
        T45 = dh_matrix(-pi/2, 0 , self.l45, axes[4])
        T56 = dh_matrix(0    , 0 , self.l56, axes[5])

        T06 = T01 @ T12 @ T23 @ T34 @ T45 @ T56
        x, y, z = list(translation_from_matrix(T06))
        roll, pitch, yaw = euler_from_matrix(T06)

        return (x, y, z, roll, pitch, yaw)

    def inverse_kinematics(self, x: float, y: float, z: float, roll: float, 
                           pitch: float, yaw: float, sol_id: int =2) -> list:
        """Inverse kinematics algorism
        
        Args:
            x (float): robot end x
            y (float): robot end y
            z (float): robot end z
            roll (float): robot end roll
            pitch (float): robot end pitch
            yaw (float): robot end yaw
            sol_id (int): inverse kinematics solve id (0-7)
        
        Returns:
            (list): list of joint angles
        """
        sol = np.zeros((6, 8))
        T06 = translation_matrix([x,y,z]) @ euler_matrix(roll, pitch, yaw)
        P05 = T06 @ np.array([[0, 0, -self.l56, 1]]).T
        
        # solve theta1
        phi1 = atan2(P05[1, 0], P05[0, 0])
        phi2 = acos(self.l34 / sqrt(P05[1, 0]*P05[1, 0] + P05[0, 0] * P05[0, 0])).real
        sol[0, 0:4] = pi / 2 + phi1 - phi2
        sol[0, 4:8] = pi / 2 + phi1 + phi2
        
        for i in range(8):
            # solve theta5
            sol[4, i] = acos((T06[0, 3] * sin(sol[0, i]) - T06[1, 3] * cos(sol[0, i]) - self.l34) / self.l56).real 
            if i % 2 == 1:
                sol[4, i] = -sol[4, i]
            
            # solve theta6
            T01 = dh_matrix(pi / 2, 0, self.l01, sol[0, i])
            T10 = np.linalg.inv(T01)
            T61 = np.linalg.inv(T10 @ T06)
            sol[5, i] = atan2(-T61[1, 2] / sin(sol[4, i]), T61[0, 2] / sin(sol[4, i])) if sol[4, i] else 0

            # solve theta3
            T45 = dh_matrix(-pi / 2, 0, self.l45, sol[4, i])
            T56 = dh_matrix(0, 0, self.l56, sol[5, i])
            T14 = T10 @ T06 @ np.linalg.inv(T56) @ np.linalg.inv(T45)
            P13 = T14 @ np.array([[0, -self.l34, 0, 1]]).T - np.array([[0, 0, 0, 1]]).T
            P13_norm = np.linalg.norm(P13)
            sol[2, i] = acos((P13_norm * P13_norm - self.l12 * self.l12 - self.l23 * self.l23) / (2 * self.l12 * self.l23)).real
            if i == 2 or i == 3 or i == 6 or i == 7:
                sol[2, i] = -sol[2, i]
            sol[1, i] = atan2(-P13[1, 0], -P13[0, 0]) - asin((self.l23 * sin(sol[2, i])) / P13_norm).real 

            # solve theta 4
            T12 = dh_matrix(0, -self.l12, 0, sol[1, i]) 
            T23 = dh_matrix(0, -self.l23, 0, sol[2, i])
            T34 = np.linalg.inv(T23) @ np.linalg.inv(T12) @ T14
            sol[3, i] = atan2(T34[1, 0], T34[0, 0]) 
            
            sol[1,i] = sol[1,i] + pi/2
            sol[2, i] = -sol[2, i]
            sol[3,i] = sol[3,i] + pi/2
        axes = []
        for i in range(6):
            axes.append(sol[i, sol_id])
        return axes


# -*-coding:utf-8 -*-
#
# Apache License, Version 2.0
#
# Copyright (c) 2023 Robottime(Beijing) Technology Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Imu sensor driver.

The imu sensor is a 6-axis inertial measurement unit (IMU) that provides
accurate acceleration and angular rate (gyroscope) measurement data for
space-constrained applications. This device also calculates and outputs
quaternions that can be used to track the absolute orientation of the sensor.

Examples:

    >>> from robodyno.components import ImuSensor
    >>> from robodyno.interfaces import CanBus
    >>> can = CanBus()
    >>> imu = ImuSensor(can)
    >>> imu.get_euler()
    (-0.6930452576214104, -0.5058813752928343, 0.33744507654497824)
    >>> can.disconnect()
    
"""

from math import pi, atan2, asin
import struct
from typing import Optional

from robodyno.interfaces import CanBus
from robodyno.components import CanBusDevice
from robodyno.components.config.model import Model


class ImuSensor(CanBusDevice):
    """Imu sensor driver.

    Attributes:
        id (int): Device id.
        type (Model): Device type.
        fw_ver (float): Firmware version.
    """

    _FC_HEARTBEAT = 0x02
    _FC_CONFIG_CAN = 0x03
    _FC_GET_ORIENTATION = 0x04
    _FC_GET_GYROSCOPE = 0x05
    _FC_GET_ACCELERATOR = 0x06
    _FC_SET_RANGES = 0x07

    def __init__(self, can: CanBus, id_: int = 0x31):
        """Initializes the imu sensor.

        Args:
            can (CanBus): Can bus instance.
            id_ (int): Device id. The default factory id is 0x31.

        Raises:
            ValueError: If the device is not a imu sensor.
        """
        super().__init__(can, id_)
        self.get_version(timeout=0.015)
        if self.type is None or self.type != Model.ROBODYNO_IMU_SENSOR:
            raise ValueError(f'Device of id {id_} is not a imu sensor')
        self._quaternion = (0, 0, 0, 1)
        self._euler = (0, 0, 0)
        self._gyro_factor = pi / 23592.96  # raw to rad/s
        self._accel_factor = 9.80665 / 16384.0  # raw to m/s^2
        self._can.subscribe(
            callback=self._heartbeat_callback,
            device_id=self.id,
            cmd_id=self._FC_HEARTBEAT,
        )

    def __del__(self):
        try:
            self._can.unsubscribe(
                callback=self._heartbeat_callback,
                device_id=self.id,
                cmd_id=self._FC_HEARTBEAT,
            )
        except KeyError:
            pass

    @property
    def quaternion(self) -> tuple:
        """Quaternion of imu sensor.

        Returns:
            (tuple): Quaternion of imu sensor.
        """
        return self._quaternion

    @property
    def euler(self) -> tuple:
        """Euler angles of imu sensor.

        Returns:
            (tuple): Euler angles of imu sensor.
        """
        return self._euler

    def _update_quaternion(self, value: tuple) -> None:
        """Updates the quaternion of imu sensor.

        Args:
            value (tuple): Quaternion of imu sensor.
        """
        self._quaternion = value
        self._euler = self.quat_to_euler(*value)

    @classmethod
    def quat_to_euler(cls, x: float, y: float, z: float, w: float) -> tuple:
        """Convert quaternion to euler angles.

        Args:
            x (float): x of quaternion.
            y (float): y of quaternion.
            z (float): z of quaternion.
            w (float): w of quaternion.

        Returns:
            (tuple): euler angles(roll, pitch, yaw).
        """
        roll = atan2(2 * (w * x + y * z), 1 - 2 * (x * x + y * y))
        pitch = asin(2 * (w * y - x * z))
        yaw = atan2(2 * (w * z + x * y), 1 - 2 * (z * z + y * y))
        return (roll, pitch, yaw)

    def _heartbeat_callback(self, data, timestamp, device_id, command_id) -> None:
        """Updates the imu sensor state."""
        del timestamp, device_id, command_id
        self._update_quaternion(struct.unpack('<eeee', data))

    def config_can_bus(
        self, new_id: int, heartbeat: int = 1000, bitrate=1000000
    ) -> None:
        """Configures the CAN bus settings.

        Args:
            new_id (int): The new device id.
            heartbeat (int): The heartbeat period in milliseconds.
            bitrate (int): The bitrate of the CAN bus. Choose from 250000,
                500000, 1000000.

        Raises:
            ValueError: If the new CAN id is not in the range of 0x01-0x3f.
        """
        if not 0x01 <= new_id <= 0x3F:
            raise ValueError('New CAN id must be in the range of 0x01-0x3f.')
        bitrate_id = {
            250000: 0,
            500000: 1,
            1000000: 2,
        }.get(bitrate, 2)
        self._can.send(
            self.id, self._FC_CONFIG_CAN, 'HHI', new_id, bitrate_id, int(heartbeat)
        )

    def set_ranges(self, gyro_range: int, accel_range: int) -> None:
        """Set gyro and accel range.

        Args:
            gyro_range: gyro range, 0: 250dps, 1: 500dps, 2: 1000dps, 3: 2000dps
            accel_range: accel range, 0: 2g, 1: 4g, 2: 8g, 3: 16g

        Raises:
            ValueError: If gyro_range or accel_range is not in the range of 0-3.
        """
        if not 0 <= gyro_range <= 3:
            raise ValueError('gyro_range must be in the range of 0-3.')
        if not 0 <= accel_range <= 3:
            raise ValueError('accel_range must be in the range of 0-3.')
        self._can.send(self.id, self._FC_SET_RANGES, 'HH', gyro_range, accel_range)

    def get_quaternion(self, timeout: Optional[float] = None) -> Optional[tuple]:
        """Reads the quaternion of imu sensor.

        Args:
            timeout (float): Timeout in seconds.

        Returns:
            (tuple | None): Quaternion of imu sensor. (x, y, z, w)
        """
        try:
            quat = self._can.get(self.id, self._FC_GET_ORIENTATION, 'eeee', timeout)
        except TimeoutError:
            return None
        self._update_quaternion(quat)
        return self._quaternion

    def get_euler(self, timeout: Optional[float] = None) -> Optional[tuple]:
        """Reads the euler angles of imu sensor.

        Args:
            timeout (float): Timeout in seconds.

        Returns:
            (tuple | None): Euler angles of imu sensor. (roll, pitch, yaw)
        """
        try:
            quat = self._can.get(self.id, self._FC_GET_ORIENTATION, 'eeee', timeout)
        except TimeoutError:
            return None
        self._update_quaternion(quat)
        return self._euler

    def get_gyro(self, timeout: Optional[float] = None) -> Optional[tuple]:
        """Reads the gyro of imu sensor.

        Args:
            timeout (float): Timeout in seconds.

        Returns:
            (tuple | None): Gyroscope of imu sensor. (gyro_x, gyro_y, gyro_z)
        """
        try:
            x, y, z, gyro_range = self._can.get(
                self.id, self._FC_GET_GYROSCOPE, 'hhhH', timeout
            )
        except TimeoutError:
            return None
        k = 2**gyro_range
        return (
            x * self._gyro_factor / k,
            y * self._gyro_factor / k,
            z * self._gyro_factor / k,
        )

    def get_accel(self, timeout: Optional[float] = None) -> Optional[tuple]:
        """Reads the accel of imu sensor.

        Args:
            timeout (float): Timeout in seconds.

        Returns:
            (tuple | None): Acceleration of imu sensor. (accel_x, accel_y, accel_z)
        """
        try:
            x, y, z, accel_range = self._can.get(
                self.id, self._FC_GET_ACCELERATOR, 'hhhH', timeout
            )
        except TimeoutError:
            return None
        k = 2**accel_range
        return (
            x * self._accel_factor / k,
            y * self._accel_factor / k,
            z * self._accel_factor / k,
        )



# -*-coding:utf-8 -*-
#
# Apache License, Version 2.0
#
# Copyright (c) 2023 Robottime(Beijing) Technology Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Robodyno interfaces for robodyno devices."""

from robodyno.interfaces.can_bus import CanBus

try:
    from robodyno.interfaces.webots import Webots

    WEBOTS_AVAILABLE = True
except RuntimeError as e:
    WEBOTS_AVAILABLE = False

from enum import Enum


class InterfaceType(Enum):
    """Robodyno supported interface types."""

    UNKNOWN = 0
    CAN_BUS = 1
    WEBOTS = 2


def get_interface_type(iface: object) -> InterfaceType:
    if isinstance(iface, CanBus):
        return InterfaceType.CAN_BUS
    elif WEBOTS_AVAILABLE and isinstance(iface, Webots):
        return InterfaceType.WEBOTS
    else:
        return InterfaceType.UNKNOWN



# -*-coding:utf-8 -*-
#
# Apache License, Version 2.0
#
# Copyright (c) 2023 Robottime(Beijing) Technology Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Provides base class for all robodyno devices in webots.

Examples:

```python
from robodyno.components import WebotsDevice
from robodyno.interfaces import Webots
webots = Webots()

class MyDevice(WebotsDevice):
    def __init__(self, webots, device_id):
        super().__init__(webots, device_id)
        self.type = Model.MY_DEVICE
        self.register()

    def __del__(self):
        self.deregister()

    def step(self):
        pass
        
    def parallel_step(self):
        pass
```
"""

from robodyno.components.config.model import Model
from robodyno.interfaces import InterfaceType, get_interface_type


class WebotsDevice(object):
    """Base class for all robodyno devices in webots.

    All robodyno devices in webots should inherit from this class.

    Attributes:
        id (int): Device id.
        type (Model): Device type.
    """

    def __init__(self, webots: object, device_id: int):
        """Initializes WebotsDevice with webots and device id.

        Args:
            webots (Webots): Webots instance.
            device_id (int): Device id.

        Raises:
            TypeError: If webots is not an instance of Webots class.
        """
        if get_interface_type(webots) != InterfaceType.WEBOTS:
            raise TypeError(
                f'interface must be an instance of Webots class, not {type(webots)}.'
            )
        self._webots = webots
        self.id = device_id
        self.type = Model.THIRD_PARTY

    def __del__(self):
        self._webots.deregister(self)

    def register(self) -> None:
        """Registers node to webots."""
        self._webots.register(self)

    def deregister(self) -> None:
        """Deregisters node from webots."""
        self._webots.deregister(self)

    def get_version(self) -> dict:
        """Gets version of the device.

        Returns:
            dict: Version of the device.
        """
        return {'main_version': 1, 'sub_version': 0, 'type': self.type}

    def parallel_step(self) -> None:
        """Parallel step function.

        This function is called in parallel with the webots simulation step.
        """
        pass

    def step(self) -> None:
        """Step function.

        This function is called after the webots simulation step.
        """
        pass



# -*-coding:utf-8 -*-
#
# Apache License, Version 2.0
#
# Copyright (c) 2023 Robottime(Beijing) Technology Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Models of Robodyno devices."""

from enum import Enum


class Model(Enum):
    """Enumerates the supported Robodyno device models."""

    ROBODYNO_PRO_P44 = 0x00
    ROBODYNO_PRO_P12 = 0x01
    ROBODYNO_PRO_P44A = 0x02
    ROBODYNO_PRO_P12A = 0x03
    ROBODYNO_PRO_H100 = 0x04
    ROBODYNO_PRO_H50 = 0x05
    ROBODYNO_PRO_B100 = 0x06
    ROBODYNO_PRO_DIRECT = 0x0F
    ROBODYNO_PLUS_H50 = 0x14
    ROBODYNO_PLUS_H100 = 0x15
    ROBODYNO_PLUS_P12 = 0x10
    ROBODYNO_PLUS_P12A = 0x12
    ROBODYNO_PLUS_DIRECT = 0x1F

    ROBODYNO_D_VAC01 = 0x61
    ROBODYNO_D_CNV02 = 0x80
    ROBODYNO_D_ADG03 = 0x20

    ROBODYNO_NANO_P100 = 0x20

    ROBODYNO_BATTERY = 0x40

    ROBODYNO_PWM_DRIVER = 0x61
    ROBODYNO_STEPPER_DRIVER = 0x62
    ROBODYNO_VACUUM_GRIPPER = 0x63
    ROBODYNO_ADAPTIVE_GRIPPER = 0x64
    ROBODYNO_FLEXIBLE_GRIPPER = 0x65
    ROBODYNO_EXB_FCTY = 0x80
    ROBODYNO_IMU_SENSOR = 0xA1
    ROBODYNO_GPS_SENSOR = 0xA2
    ROBODYNO_LED_DRIVER = 0xA3

    THIRD_PARTY = 0xFF

    @classmethod
    def _missing_(cls, value):
        return cls.THIRD_PARTY



# -*-coding:utf-8 -*-
#
# Apache License, Version 2.0
#
# Copyright (c) 2023 Robottime(Beijing) Technology Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provides parameter types for the CLI."""

import click
from rich.prompt import PromptBase, DefaultType, InvalidResponse
from rich.text import Text


class DeviceIdListParamType(click.ParamType):
    """A click parameter type for device ID list."""

    name = 'device_id_list'

    def convert(self, value, param, ctx):
        if isinstance(value, int):
            return (value,)

        try:
            id_set = set()
            for id_str in value.split(','):
                splited = id_str.split('-')
                if len(splited) == 1:
                    id_set.add(int(splited[0], 0))
                elif len(splited) == 2:
                    id_set.update(range(int(splited[0], 0), int(splited[1], 0) + 1))
                else:
                    raise ValueError
            return tuple(id_set)
        except ValueError:
            self.fail(f'{value!r} is not a valid id list', param, ctx)


ID_LIST = DeviceIdListParamType()


class BasedIntParamType(click.ParamType):
    """A click parameter type for based integer."""

    name = 'integer'

    def convert(self, value, param, ctx):
        if isinstance(value, int):
            return value

        try:
            return int(value, 0)
        except ValueError:
            self.fail(f'{value!r} is not a valid integer', param, ctx)


BASED_INT = BasedIntParamType()


class BasedIntPrompt(PromptBase[int]):
    """A based integer prompt.

    Example:
        >>> id = BasedIntPrompt.ask("Please enter a device id", default=0x01)
    """

    response_type = int
    validate_error_message = '[prompt.invalid]Please enter a valid integer number'

    def render_default(self, default: DefaultType) -> Text:
        """Render the default as (y) or (n) rather than True/False."""
        return Text(f'(0x{default:02X})', style='prompt.default')

    def process_response(self, value: str) -> int:
        """Convert choices to a int."""
        value = value.strip().lower()
        try:
            value = int(value, 0)
        except ValueError as error:
            raise InvalidResponse(self.validate_error_message) from error
        return value



# -*-coding:utf-8 -*-
#
# Apache License, Version 2.0
#
# Copyright (c) 2023 Robottime(Beijing) Technology Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base class for all robodyno devices which use can bus as communication

All robodyno devices which use can bus as communication interface should
inherit this class.

Examples:

    >>> from robodyno.components import CanBusDevice
    >>> from robodyno.interfaces import CanBus
    >>> can_bus = CanBus()
    >>> device = CanBusDevice(can_bus, 0x01)
    >>> device.get_version()
    {'main_version': 1, 'sub_version': 0, 'type': <Model.THIRD_PARTY: 255>}
    >>> device.fw_ver
    1.0
    >>> device.type
    <Model.THIRD_PARTY: 255>
    >>> can_bus.disconnect()
"""

from typing import Optional

from robodyno.components.config.model import Model
from robodyno.interfaces import InterfaceType, get_interface_type


class CanBusDevice(object):
    """Base class for all robodyno devices which use can bus as communication
    interface.

    This class provides basic functions for all robodyno devices which use can
    bus as communication interface. It provides functions to get device
    firmware version and device type.

    Attributes:
        id (int): Device id.
        type (Model): Device type.
        fw_ver (float): Device firmware version.
    """

    _CMD_GET_VERSION = 0x01

    def __init__(self, can, device_id):
        """Init device with interface and id.

        Args:
            can (CanBus): Can bus interface.
            device_id (int): Device id.

        Raises:
            ValueError: If device id is not in range [0x01, 0x40).
            TypeError: If can is not CanBus.
        """
        if not 0x01 <= device_id < 0x40:
            raise ValueError('Device id should be in range [0x01, 0x40).')
        if get_interface_type(can) != InterfaceType.CAN_BUS:
            raise TypeError('CanBusDevice can only be initialized with CanBus.')
        self._can = can
        self.id = device_id
        self.fw_ver = None
        self.type = Model.THIRD_PARTY

    def get_version(self, timeout: Optional[float] = None) -> Optional[dict]:
        """Get device firmware version and type.

        Args:
            timeout (float): Timeout in seconds.

        Returns:
            (dict | None): Device firmware version and type.
        """
        try:
            main_ver, sub_ver, type_ = self._can.get(
                self.id, self._CMD_GET_VERSION, 'HHI', timeout
            )
        except TimeoutError:
            return None
        self.fw_ver = float(f'{main_ver}.{sub_ver}')
        self.type = Model(type_)
        return {
            'main_version': main_ver,
            'sub_version': sub_ver,
            'type': Model(type_),
        }



# -*-coding:utf-8 -*-
#
# The MIT License (MIT)
#
# Copyright (c) 2023 Robottime(Beijing) Technology Co., Ltd
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""This module provides a class for controlling Robodyno Four DoF Scara Robot.

The FourDoFScara class provided by this module is used to control Robodyno Four DoF Scara Robot
through the CAN bus. It provides methods for setting Four DoF Scara Robot parameters, reading Four DoF Scara Robot
states, and controlling the Four DoF Scara Robot to run in different space.

Examples:

    >>> from robodyno.components import Motor
    >>> from robodyno.interfaces import CanBus
    >>> from robodyno.robots.four_dof_scara_robot import FourDoFScara
    >>> can = CanBus()
    >>> class MyScara(FourDoFScara):
    >>>     def __init__(self):
    >>>         M1 = Motor(can, 0x10, 'ROBODYNO_PRO_44')
    >>>         M2 = Motor(can, 0x11, 'ROBODYNO_PRO_44')
    >>>         M3 = Motor(can, 0x12, 'ROBODYNO_PRO_12')
    >>>         M4 = Motor(can, 0x13, 'ROBODYNO_PRO_12')
    >>>     super().__init__(M1, M2, M3, M4, 0.06, 0.185, 0.185, 0.255, 0.01)
    >>> arm = MyScara()
    >>> arm.init()
    >>> arm.get_joints_poses()
    [0.0, 0.0, 0.0, 0.0]
    >>> can.disconnect()
"""

import time
from math import pi, cos, sin, sqrt, atan2
from ..utils.interpolations import linear_interpolation

class FourDoFScara(object):
    """Controls Robodyno Four DoF Scara Robot through the CAN bus.
    
    Attributes:
        joints (list): list of 4 joint motors
        l12 (float): link from joint 1 to joint 2  (m)
        l23 (float): link from joint 2 to joint 3  (m)
        l34 (float): link from joint 3 to joint 4  (m)
        l04 (float): link from world to joint 4 longitudinal distance (m)
        screw_lead (float): screw lead (m)
        end_effector (object): end effector object

    """
    def __init__(self, j1, j2, j3, j4, l12: float, l23: float, l34: float, 
                 l04: float, screw_lead: float, end_effector: object = None):
        """Initializes robot with joints and links
        
        Args:
            j1 (Motor): shoulder_motor.
            j2 (Motor): upperarm_motor.
            j3 (Motor): forearm_motor.
            j4 (Motor): hand_motor.
            l12 (float): link from joint 1 to joint 2  (m)
            l23 (float): link from joint 2 to joint 3  (m)
            l34 (float): link from joint 3 to joint 4  (m)
            l04 (float): link from world to joint 4 longitudinal distance (m)
            screw_lead (float): screw lead (m)
            end_effector (object): end effector object
        """
        self.joints = [j1, j2, j3, j4]
        self.l12 = l12
        self.l23 = l23
        self.l34 = l34
        self.l04 = l04
        self.screw_lead = screw_lead
        self.end_effector = end_effector

        for m in self.joints:
            m.position_filter_mode(8)
        # self.motors[0].position_traj_mode(10,5,5)
        self._axes_poses = [0 for i in range(4)]
        self._axes_zeros = [0 for i in range(4)]

    def get_joints_poses(self) -> list:
        """Read joints positions to a list.
        
        Returns:
            (list): a list of 4 joint positions
            
        Raises:
            RuntimeError: If the motor Joint is invalid.
        """
        poses = []
        for i in range(4):
            pos = None
            count = 0
            while pos is None:
                if count > 4:
                    raise RuntimeError('Filed to get position of Joint {}'.format(i+1))
                count += 1
                pos = self.joints[i].get_pos(0.2)
            poses.append(pos - self._axes_zeros[i])
        return poses

    def enable(self) -> None:
        """enable joints motors"""
        for i in range(4):
            self.joints[i].enable()

    def disable(self) -> None:
        """disable joints motors"""
        for i in range(4):
            self.joints[i].disable()

    def init(self, axes_poses: list = [0 for i in range(4)]) -> None:
        """Calibrate robot motors with given axes poses.
        
        Args:
            axes_poses (list): list of 4 axes poses(rad)
        """
        self._axes_poses = axes_poses.copy()
        self._axes_zeros = [0 for i in range(4)]
        cur_poses = self.get_joints_poses()
        for i in range(4):
            self._axes_zeros[i] = cur_poses[i] - self._axes_poses[i]
            self._axes_poses[i] = 0

    def set_joint_pos(self, id: int, pos: float) -> None:
        """Set joint angle with joint id and position
        
        Args:
            id (int): joint id (0-3)
            pos (float): joint target position(rad)
        """
        self.joints[id].set_pos(pos + self._axes_zeros[id])
        self._axes_poses[id] = pos

    def joint_space_interpolated_motion(self, target: list, 
                                        speeds: list = [None for i in range(4)], 
                                        duration: float = 0) -> None:
        """Robot interpolated motion in joint space.
        
        Args:
            target (list): iterable of 4 joints target angle(rad)
            speeds (list): iterable of 4 joints motion speed(rad/s)
            duration (float): default motion duration(s)
        """
        interpolation = []
        joint_poses = self._axes_poses.copy()
        for i in range(4):
            interpolation.append(linear_interpolation(joint_poses[i], target[i], speed=speeds[i], duration=duration))
        update_flag = True
        while update_flag:
            update_flag = False
            for i in range(4):
                pos = next(interpolation[i], None)
                if pos is not None:
                    self.set_joint_pos(i, pos)
                    update_flag = True
            time.sleep(0.05)

    def home(self, duration: float = 5) -> None:
        """Move back to zero position.
        
        Args:
            duration (float): motion duration(s)
        """
        self.joint_space_interpolated_motion((0,0,0,0), duration = duration)

    def cartesian_space_interpolated_motion(self, x: float, y: float, z: float, yaw: float, x_speed: float = None, 
                                            y_speed: float = None, z_speed: float = None, yaw_speed: float = None, 
                                            hand_coordinate: bool = 1, duration: float = 0) -> None:
        """Robot Interpolated motion in cartesian space.
        
        Args:
            x (float): target robot end x
            y (float): target robot end y
            z (float): target robot end z
            yaw (float): target robot end yaw
            x_speed (float): speed alone X dimension(m/s)
            y_speed (float): speed alone Y dimension(m/s)
            z_speed (float): speed alone Z dimension(m/s)
            yaw_speed (float): rotation speed on Z axis(rad/s)
            hand_coordinate (bool): 0 is left hand coordinate system
                             1 is right hand coordinate system
            duration (float): default motion duration(s)
        """
        current_pose = self.forward_kinematics(self._axes_poses.copy())
        current_x = current_pose[0]
        current_y = current_pose[1]
        current_z = current_pose[2]
        current_yaw = current_pose[3]

        x_interpolation = linear_interpolation(current_x, x, x_speed, duration)
        y_interpolation = linear_interpolation(current_y, y, y_speed, duration)
        z_interpolation = linear_interpolation(current_z, z, z_speed, duration)
        yaw_interpolation = linear_interpolation(current_yaw, yaw, yaw_speed, duration)
        
        while True:
            temp_x = next(x_interpolation, x)
            temp_y = next(y_interpolation, y)
            temp_z = next(z_interpolation, z)
            temp_yaw = next(yaw_interpolation, yaw)

            axes = self.inverse_kinematics(temp_x, temp_y, temp_z, temp_yaw , hand_coordinate)
            for i in range(4):
                self.set_joint_pos(i, axes[i])
            if temp_x == x and temp_y == y and temp_z == z and temp_yaw == yaw:
                break
            time.sleep(0.05)

    def forward_kinematics(self, angles: list) -> tuple:
        """Forward kinematics algorism
        
        Args:
            angles (list): list of 4 joint angles(rad)
        
        Returns:
            (tuple): (x, y, z, yaw) tuples of 3 axis position and 1 axis posture
        """
        x = self.l23 * cos(angles[1]) + self.l34 * cos(angles[1] + angles[2]) + self.l12
        y = self.l23 * sin(angles[1]) + self.l34 * sin(angles[1] + angles[2])
        z = angles[0] * self.screw_lead / (2 * pi) + self.l04
        yaw = -angles[3] + angles[1] + angles[2]

        return (x, y, z, yaw)
    
    def inverse_kinematics(self, x: float, y: float, z: float, 
                           yaw: float, hand_coordinate: bool = 1) -> list:
        """inverse kinematics algorism
        
        Args:
            x (float): robot end x
            y (float): robot end y
            z (float): robot end z
            yaw (float): robot end yaw
            hand_coordinate (bool): 0 is left hand coordinate system
                             1 is right hand coordinate system
        
        Returns:
            (list): list of 4 joint angles
            
        Raises:
            ValueError: If the Scara Pose not in range.
        """
        angle = [0, 0, 0, 0]
        x -= self.l12
        z -= self.l04
        
        c3 = (x*x + y*y - self.l23*self.l23 - self.l34*self.l34) / (2.0 * self.l23 * self.l34)
        temp = 1 - c3 * c3
        try:
            if(hand_coordinate == 1):
                s3 = sqrt(temp)  # right
            elif(hand_coordinate == 0):
                s3 = -sqrt(temp) # left
        except ValueError:
            print("Pose not in range! Choose other Pose that is not a singularity")
            return self.get_joints_poses()
         
        angle[2] = atan2(s3, c3)
         
        alpha = atan2(y, x)
        r = sqrt(x*x + y*y)
        sin_beta = self.l34 * sin(angle[2]) / r
        cos_beta = (self.l23 + self.l34 * cos(angle[2])) / r
        beta = atan2(sin_beta, cos_beta)
        angle[1] = alpha - beta
        
        angle[0] = (2*pi * z) / self.screw_lead
        angle[3] = angle[1] + angle[2] - yaw
         
        return angle
        



import sys
if sys.version_info < (3, 10):
    from importlib_metadata import entry_points
else:
    from importlib.metadata import entry_points
thismodule = sys.modules[__name__]

display_eps = entry_points(group='robodyno.robots')
for ep in display_eps:
    setattr(thismodule, ep.name, ep.load())



# -*-coding:utf-8 -*-
#
# The MIT License (MIT)
#
# Copyright (c) 2023 Robottime(Beijing) Technology Co., Ltd
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""This module provides a class for controlling Robodyno Three DoF Cartesian Robot.

The ThreeDoFCartesian class provided by this module is used to control Robodyno Three DoF Cartesian Robot
through the CAN bus. It provides methods for setting Three DoF Cartesian Robot parameters, reading Three DoF Cartesian Robot
states, and controlling the Three DoF Cartesian Robot to run in different space.

Examples:

    >>> from robodyno.components import Motor
    >>> from robodyno.interfaces import CanBus
    >>> from robodyno.robots.three_dof_cartesian_robot import ThreeDoFCartesian
    >>> can = CanBus()
    >>> class MyCartesian(ThreeDoFCartesian):
    >>>     def __init__(self):
    >>>         M1 = Motor(can, 0x10, 'ROBODYNO_PRO_44')
    >>>         M2 = Motor(can, 0x11, 'ROBODYNO_PRO_44')
    >>>         M3 = Motor(can, 0x12, 'ROBODYNO_PRO_44')
    >>>     super().__init__(M1, M2, M3, 0.0102, 0.0102, 0.0102)
    >>> arm = MyCartesian()
    >>> arm.init()
    >>> arm.get_joints_poses()
    [0.0, 0.0, 0.0]
    >>> can.disconnect()
"""

import time
from math import pi
from ..utils.interpolations import linear_interpolation

class ThreeDoFCartesian(object):
    """Controls Robodyno Three Dof Cartesian Robot through the CAN bus.
    
    Attributes:
        joints (list): list of 3 joint motors
        screw_lead_x (float): The lead screw on the x axis (m)
        screw_lead_y (float): The lead screw on the y axis (m)
        screw_lead_z (float): The lead screw on the z axis (m)
        end_effector (object): end effector object
    """
    def __init__(self, j1, j2, j3, screw_lead_x: float, screw_lead_y: float, 
                 screw_lead_z: float, end_effector: object = None):
        """Initializes robot with joints and links
        
        Args:
            j1 (Motor): cartesian x axis joint.
            j2 (Motor): cartesian y axis joint.
            j3 (Motor): cartesian z axis joint.
            screw_lead_x (float): The lead screw on the x axis (m)
            screw_lead_y (float): The lead screw on the y axis (m)
            screw_lead_z (float): The lead screw on the z axis (m)
            end_effector (object): end effector object
        """
        self.joints = [j1, j2, j3]
        self.screw_lead_x = screw_lead_x
        self.screw_lead_y = screw_lead_y
        self.screw_lead_z = screw_lead_z
        self.end_effector = end_effector

        for m in self.joints:
          m.position_filter_mode(8)
        self._axes_poses = [0 for i in range(3)]
        self._axes_zeros = [0 for i in range(3)]
        
    def get_joints_pose(self) -> list:
        """Read joints position to a list.
        
        Returns:
            (list): a list of 3 joint positions
            
        Raises:
            RuntimeError: If the motor Joint is invalid.
        """
        poses = []
        for i in range(3):
            pos = None
            count = 0
            while pos is None:
                if count > 3:
                    raise RuntimeError('Filed to get position of Joint {}'.format(i+1))
                count += 1
                pos = self.joints[i].get_pos(0.2)
            poses.append(pos - self._axes_zeros[i])
        return poses
    
    def enable(self) -> None:
        """enable joints motors"""
        for i in range(3):
            self.joints[i].enable()

    def disable(self) -> None:
        """disable joints motors"""
        for i in range(3):
            self.joints[i].disable()
            
    def init (self, axes_poses: list = [0 for i in range(3)]) -> None:
        """Calibrate robot motors with given axes poses.
        
        Args:
            axes_poses (list): list of 3 axes poses(rad)
        """
        self._axes_poses = axes_poses.copy()
        self._axes_zeros = [0 for i in range(3)]
        cur_poses = self.get_joints_pose()
        for i in range(3):
            self._axes_zeros[i] = cur_poses[i] - self._axes_poses[i]
            self._axes_poses[i] = 0

    def set_joint_pos(self, id: int, pos: float) -> None:
        """Set joint angle with joint and position.
        
        Args:
            id (int): joint id (0-2)
            pos (float): joint target position(red)
        """
        self.joints[id].set_pos(pos + self._axes_zeros[id])
        self._axes_poses[id] = pos
        
    def joint_space_interpolated_motion(self, target: list, 
                                        speeds: list = [0 for i in range(3)], 
                                        duration: float = 0) -> None:
        """Robot interpolated motion in joint space.
        
        Args:
            target (list): iterable of 3 joints target angle(rad)
            speeds (list): iterable of 3 joints motion speed(rad/s)
            duration (float): default motion duration(s)
        """
        interpolation = []
        joint_poses = self._axes_poses.copy()
        for i in range(3):
            interpolation.append(linear_interpolation(joint_poses[i], target[i], speed=speeds[i], duration=duration))
        update_flag = True
        while update_flag:
            update_flag = False
            for i in range(3):
                pos = next(interpolation[i], None)
                if pos is not None:
                    self.set_joint_pos(i, pos)
                    update_flag = True
            time.sleep(0.05)

    def home(self, duration: float = 5) -> None:
        """Move back to zero position.
        
        Args:
            duration (float): motion duration(s)
        """
        self.joint_space_interpolated_motion((0,0,0), duration = duration)

    def cartesian_space_interpolated_motion(self, x: float, y: float, z: float, x_speed: float = None, 
                                            y_speed: float = None, z_speed: float = None, duration: float = 0) -> None:
        """Robot Interpolated motion in cartesian space.
        
        Args:
            x (float): target robot end x
            y (float): target robot end y
            z (float): target robot end z
            x_speed (float): speed alone X dimension(m/s)
            y_speed (float): speed alone Y dimension(m/s)
            z_speed (float): speed alone Z dimension(m/s)
            duration (float): default motion duration(s)
        """
        current_pose = self.forward_kinematics(self._axes_poses.copy())
        current_x = current_pose[0]
        current_y = current_pose[1]
        current_z = current_pose[2]

        x_interpolation = linear_interpolation(current_x, x, x_speed, duration)
        y_interpolation = linear_interpolation(current_y, y, y_speed, duration)
        z_interpolation = linear_interpolation(current_z, z, z_speed, duration)
        
        while True:
            temp_x = next(x_interpolation, x)
            temp_y = next(y_interpolation, y)
            temp_z = next(z_interpolation, z)

            axes = self.inverse_kinematics(temp_x, temp_y, temp_z)
            for i in range(3):
                self.set_joint_pos(i, axes[i])
            if temp_x == x and temp_y == y and temp_z == z:
                break
            time.sleep(0.05)
            
    def forward_kinematics(self, angles: list) -> tuple:
        """Forward kinematics algorism
        
        Args:
            angles (list): list of 3 joint angles(rad)
        
        Returns:
            (tuple): (x, y, z) tuples of 3 axis position
        """
        x = -angles[0] * self.screw_lead_x / (2 * pi)
        y = -angles[1] * self.screw_lead_y / (2 * pi)
        z =  angles[2] * self.screw_lead_z / (2 * pi)
        
        return (x, y, z)
        

    def inverse_kinematics(self, x: float, y: float, z: float) -> list:
        """inverse kinematics algorism
        
        Args:
            x (float): robot end x
            y (float): robot end y
            z (float): robot end z
        
        Returns:
            (list): list of joint angles
        """
        theta = [0,0,0]
        theta[0] = (2*pi * -x) / self.screw_lead_x
        theta[1] = (2*pi * -y) / self.screw_lead_y
        theta[2] = (2*pi *  z) / self.screw_lead_z
        return theta
        



# -*-coding:utf-8 -*-
#
# The MIT License (MIT)
#
# Copyright (c) 2023 Robottime(Beijing) Technology Co., Ltd
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""This module provides a class for controlling Robodyno Three DoF Delta Robot.

The ThreeDoFDelta class provided by this module is used to control Robodyno Three DoF Delta Robot
through the CAN bus. It provides methods for setting Three DoF Delta Robot parameters, reading Three DoF Delta Robot
states, and controlling the Three DoF Delta Robot to run in different space.

Examples:

    >>> from robodyno.components import Motor
    >>> from robodyno.interfaces import CanBus
    >>> from robodyno.robots.three_dof_delta_robot import ThreeDoFDelta
    >>> can = CanBus()
    >>> class MyDelta(ThreeDoFDelta):
    >>>     def __init__(self):
    >>>         M1 = Motor(can, 0x10, 'ROBODYNO_PRO_44')
    >>>         M2 = Motor(can, 0x11, 'ROBODYNO_PRO_44')
    >>>         M3 = Motor(can, 0x12, 'ROBODYNO_PRO_44')
    >>>     super().__init__(M1, M2, M3, 0.12864, 0.3, 0.13856, 0.025)
    >>> arm = MyDelta()
    >>> arm.init()
    >>> arm.get_joints_poses()
    [0.0, 0.0, 0.0]
    >>> can.disconnect()
"""
"""three_dof_delta_robot.py
Time    :   2022/10/09
Author  :   ryan 
Version :   1.0
Contact :   ryanzhang@163.com
License :   (C)Copyright 2022, robottime / robodyno

3DoF Delta Robot Drive

  Typical usage example:

  from robodyno.robots.three_dof_delta_robot import ThreeDoFDelta
  robot = ThreeDoFDelta(
    j1 = delta_motor1,
    j2 = delta_motor2,
    j3 = delta_motor3,
    l1 = 0.12864,
    l2 = 0.3,
    r1 = 0.13856,
    r2 = 0.025,
    end_effector = None
  )
  
"""
import time
import numpy as np
from math import pi, cos, sin, sqrt, atan
from ..utils.interpolations import linear_interpolation


class ThreeDoFDelta(object):
    """Controls Robodyno Three DoF Delta Robot through the CAN bus.
    
    Attributes:
        joints (list): list of 3 joint motors
        l1 (float): link is the active arm (m)
        l2 (float): link is the slave arm (m) 
        r1 (float): fixed platform radius (m)
        r2 (float): motion platform radius (m)
        end_effector (object): end effector object
    """
    def __init__(self, j1, j2, j3, l1: float, l2: float, 
                 r1: float, r2: float, end_effector: object = None):
        """Initializes robot with joints and links
        
        Args:
            j1 (Motor): delta joint 1.
            j2 (Motor): delta joint 2.
            j3 (Motor): delta joint 3.
            l1 (float): link is the active arm (m)
            l2 (float): link is the slave arm (m) 
            r1 (float): fixed platform radius (m)
            r2 (float): motion platform radius (m)
            end_effector (object): end effector object
        """
        self.joints = [j1, j2, j3]
        self.l1 = l1
        self.l2 = l2
        self.r1 = r1
        self.r2 = r2
        self.end_effector = end_effector

        for m in self.joints:
            m.position_filter_mode(8)
        self._axes_poses = [0 for i in range(3)]
        self._axes_zeros = [0 for i in range(3)]
        self._link_deviation = 0.08 # 4.63°

    def get_joints_poses(self) -> list:
        """Read joints positions to a list.

        Returns:
            (list): a list of 3 joint positions
            
        Raises:
            RuntimeError: If the motor Joint is invalid.
        """
        poses = []
        for i in range(3):
            pos = None
            count = 0
            while pos is None:
                if count > 3:
                    raise RuntimeError('Filed to get position of Joint {}'.format(i+1))
                count += 1
                pos = self.joints[i].get_pos(0.2)
            poses.append(pos - self._axes_zeros[i])
        return poses

    def enable(self) -> None:
        """enable joints motors"""
        for i in range(3):
            self.joints[i].enable()

    def disable(self) -> None:
        """disable joints motors"""
        for i in range(3):
            self.joints[i].disable()

    def init(self, axes_poses: list = [0 for i in range(3)]) -> None:
        """Calibrate robot motors with given axes poses.
        
        Args:
            axes_poses (list): list of 3 axes poses(rad)
        """
        self._axes_poses = axes_poses.copy()
        self._axes_zeros = [0 for i in range(3)]
        cur_poses = self.get_joints_poses()
        for i in range(3):
            self._axes_zeros[i] = cur_poses[i] - self._axes_poses[i]
            self._axes_poses[i] = 0

    def set_joint_pos(self, id: int, pos: float) -> None:
        """Set joint angle with joint id and position
        
        Args:
            id (int): joint id (0-2)
            pos (float): joint target position(rad)
        """
        self.joints[id].set_pos(pos + self._axes_zeros[id])
        self._axes_poses[id] = pos

    def joint_space_interpolated_motion(self, target: list, 
                                        speeds: list = [None for i in range(3)], 
                                        duration: float = 0) -> None:
        """Robot interpolated motion in joint space.
        
        Args:
            target (list): iterable of 3 joints target angle(rad)
            speeds (list): iterable of 3 joints motion speed(rad/s)
            duration (float): default motion duration(s)
        """
        interpolation = []
        joint_poses = self._axes_poses.copy()
        for i in range(3):
            interpolation.append(linear_interpolation(joint_poses[i], target[i], speed=speeds[i], duration=duration))
        update_flag = True
        while update_flag:
            update_flag = False
            for i in range(3):
                pos = next(interpolation[i], None)
                if pos is not None:
                    self.set_joint_pos(i, pos)
                    update_flag = True
            time.sleep(0.05)
    
    def home(self, duration: float = 5) -> None:
        """Move back to zero position.
        
        Args:
            duration (float): motion duration(s)
        """
        self.joint_space_interpolated_motion((0,0,0), duration = duration)

    def cartesian_space_interpolated_motion(self, x: float, y: float, z: float, x_speed: float = None, 
                                            y_speed: float = None, z_speed: float = None, duration: float = 0) -> None:
        """Robot Interpolated motion in cartesian space.
        
        Args:
            x (float): target robot end x
            y (float): target robot end y
            z (float): target robot end z
            x_speed (float): speed alone X dimension(m/s)
            y_speed (float): speed alone Y dimension(m/s)
            z_speed (float): speed alone Z dimension(m/s)
            duration (float): default motion duration(s)
        """
        current_pose = self.forward_kinematics(self._axes_poses.copy())
        current_x = current_pose[0]
        current_y = current_pose[1]
        current_z = current_pose[2]

        x_interpolation = linear_interpolation(current_x, x, x_speed, duration)
        y_interpolation = linear_interpolation(current_y, y, y_speed, duration)
        z_interpolation = linear_interpolation(current_z, z, z_speed, duration)
        
        while True:
            temp_x = next(x_interpolation, x)
            temp_y = next(y_interpolation, y)
            temp_z = next(z_interpolation, z)

            axes = self.inverse_kinematics(temp_x, temp_y, temp_z)
            for i in range(3):
                self.set_joint_pos(i, axes[i])
            if temp_x == x and temp_y == y and temp_z == z:
                break
            time.sleep(0.05)

    def forward_kinematics(self, angles: list) -> tuple:
        """Forward kinematics algorism
        
        Args:
            angles (list): list of 3 joint angles(rad)
        
        Returns:
            (tuple): (x, y, z) tuples of 3 axis position
        """
        cos_0 = cos(0) # 0°
        sin_0 = sin(0)
        cos_120 = cos(2 * pi / 3) # 120°
        sin_120 = sin(2 * pi / 3)
        cos_240 = cos(4 * pi / 3) # 240°
        sin_240 = sin(4 * pi / 3)
        
        theta1 = angles[0] + self._link_deviation  
        theta2 = angles[1] + self._link_deviation 
        theta3 = angles[2] + self._link_deviation 

        O_C1 = np.array([self.r1 * cos_0  , self.r1 * sin_0  , 0])
        O_C2 = np.array([self.r1 * cos_120, self.r1 * sin_120, 0])
        O_C3 = np.array([self.r1 * cos_240, self.r1 * sin_240, 0])

        C1_B1 = np.array([-self.l1*sin(theta1)*cos_0  , -self.l1*sin(theta1)*sin_0  , -self.l1*cos(theta1)])
        C2_B2 = np.array([-self.l1*sin(theta2)*cos_120, -self.l1*sin(theta2)*sin_120, -self.l1*cos(theta2)])
        C3_B3 = np.array([-self.l1*sin(theta3)*cos_240, -self.l1*sin(theta3)*sin_240, -self.l1*cos(theta3)])

        A1_P = np.array([-self.r2*cos_0  , -self.r2*sin_0  , 0])
        A2_P = np.array([-self.r2*cos_120, -self.r2*sin_120, 0])
        A3_P = np.array([-self.r2*cos_240, -self.r2*sin_240, 0])

        O_D1 = O_C1 + C1_B1 + A1_P
        O_D2 = O_C2 + C2_B2 + A2_P
        O_D3 = O_C3 + C3_B3 + A3_P
        
        D2_D1 = O_D1 - O_D2
        D2_D3 = O_D3 - O_D2
        D3_D1 = O_D1 - O_D3
        D3_D2 = O_D2 - O_D3

        a = np.linalg.norm(D2_D1)
        b = np.linalg.norm(D2_D3)
        c = np.linalg.norm(D3_D1)

        p = (a + b + c) / 2
        S = sqrt(p * (p-a) * (p-b) * (p-c))
        D2_E_norm = a * b * c / (4 * S)

        D2_F_norm = b / 2
        FE_norm = np.linalg.norm(sqrt(D2_E_norm * D2_E_norm - D2_F_norm * D2_F_norm))

        nF_E1 = np.cross(np.cross(D2_D1, D2_D3), D3_D2) / (a * b * b)
        nF_E = nF_E1 / np.linalg.norm(nF_E1)

        nE_P1 = np.cross(D2_D1, D2_D3) / (a * b)
        nE_P = nE_P1 / np.linalg.norm(nE_P1)

        D1_P_norm = self.l2
        D1_E_norm = a * b * c / (4 * S)

        EP_norm = sqrt(D1_P_norm * D1_P_norm - D1_E_norm * D1_E_norm)
        EP = EP_norm * nE_P

        OF = (O_D2 + O_D3) / 2
        FE = FE_norm *nF_E
        OE = OF + FE

        OP = OE + EP

        x = OP[0]
        y = OP[1]
        z = OP[2]

        return (x, y, z)

    def inverse_kinematics(self, x: float, y: float, z: float) -> list:
        """inverse kinematics algorism
        
        Args:
            x (float): robot end x
            y (float): robot end y
            z (float): robot end z
        
        Returns:
            (list): list of joint angles
            
        Raises:
            ValueError: If the Delta robot Pose not in range.
        """
        angles = [0, 0, 0]
        m = x*x + y*y + z*z + (self.r1-self.r2)*(self.r1-self.r2) + self.l1*self.l1 - self.l2*self.l2 
        A = [(m - 2 * x * (self.r1 - self.r2)) / (2 * self.l1) - (self.r1 - self.r2 - x),
             (m + (self.r1 - self.r2) * (x - sqrt(3) * y)) / (self.l1) - 2 * (self.r1 - self.r2) - (x - sqrt(3) * y),
             (m + (self.r1 - self.r2) * (x + sqrt(3) * y)) / (self.l1) - 2 * (self.r1 - self.r2) - (x + sqrt(3) * y)]

        B = [2 * z,
             4 * z,
             4 * z]
        C = [(m - 2 * x * (self.r1 - self.r2)) / (2 * self.l1) + (self.r1 - self.r2 - x),
             (m + (self.r1 - self.r2) * (x - sqrt(3) * y)) / (self.l1) + 2 * (self.r1 - self.r2) + (x - sqrt(3) * y),
             (m + (self.r1 - self.r2) * (x + sqrt(3) * y)) / (self.l1) + 2 * (self.r1 - self.r2) + (x + sqrt(3) * y)]
        
        delta = [0,0,0]
        for i in range(3):
            delta[i] = B[i] * B[i] - 4 * A[i] * C[i]
            try:
                t = (-B[i] - sqrt(delta[i])) / (2 * A[i])
                angles[i] = 2 * atan(t) - (pi/2 + self._link_deviation)
            except ValueError:
                print('Pose not in range! Choose other Pose that is not a singularity')
                return self.get_joints_poses()

        return angles




# -*-coding:utf-8 -*-
#
# Apache License, Version 2.0
#
# Copyright (c) 2023 Robottime(Beijing) Technology Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pwm develop board driver.

The pwm develop board has a 8-bit pwm output or servo output on the signal pin.
It can be used to control the servo or pwm device.

Examples:

    >>> from robodyno.components import PwmDriver
    >>> from robodyno.interfaces import CanBus
    >>> can = CanBus()
    >>> pwm = PwmDriver(can)
    >>> pwm.set_pwm(100)
    >>> pwm.set_servo(100)
    >>> can.disconnect()
"""

from robodyno.interfaces import CanBus
from robodyno.components import CanBusDevice
from robodyno.components.config.model import Model


class PwmDriver(CanBusDevice):
    """Pwm develop board driver.

    Attributes:
        id (int): Device id.
        type (Model): Device type.
        fw_ver (float): Firmware version.
    """

    _CMD_SET_NODE_ID = 0x02
    _CMD_SET_SERVO = 0x03
    _CMD_SET_PWM = 0x04

    def __init__(self, can: CanBus, id_: int = 0x21):
        """Initializes the pwm develop board.

        Args:
            can (CanBus): Can bus instance.
            id_ (int): Device id.

        Raises:
            ValueError: If the device is not a pwm develop board.
        """
        super().__init__(can, id_)
        self.get_version(timeout=0.015)
        if self.type is None or self.type != Model.ROBODYNO_D_VAC01:
            raise ValueError('The device is not a pwm develop board.')
        self.off()
        self._pwm = 0
        self._servo = None

    @property
    def pwm(self) -> int:
        """Gets the pwm value.

        Returns:
            (int): 8-bit pwm value. The range is 0-255. None if the device is in
                servo mode.
        """
        return self._pwm

    @property
    def servo_pos(self) -> int:
        """Gets the servo position.

        Returns:
            (int): Position value. The range is 0-255 (0-180 degree). None if the
                device is in pwm mode.
        """
        return self._servo

    def config_can_bus(self, new_id: int) -> None:
        """Configures the can bus.

        Args:
            new_id (int): New device id.

        Raises:
            ValueError: If the new id is not in the range of 0x01-0x3f.
        """
        if not 0x01 <= new_id <= 0x3F:
            raise ValueError('New CAN id must be in the range of 0x01-0x3f.')
        self._can.send(self.id, self._CMD_SET_NODE_ID, 'B', new_id)

    def on(self) -> None:
        """Turns on the pwm output."""
        self.set_pwm(255)

    def off(self) -> None:
        """Turns off the pwm output."""
        self.set_pwm(0)

    def set_servo(self, pos: int) -> None:
        """Sets the servo position.

        Args:
            pos (int): Position value. The range is 0-255 (0-180 degree).

        Raises:
            ValueError: If the position is not in the range of 0-255.
        """
        if not 0 <= pos <= 255:
            raise ValueError('Servo position must be in the range of 0-255.')
        self._servo = pos
        self._pwm = None
        self._can.send(self.id, self._CMD_SET_SERVO, 'B', pos)

    def set_pwm(self, pwm: int) -> None:
        """Sets the pwm value.

        Args:
            pwm (int): 8-bit pwm value. The range is 0-255.

        Raises:
            ValueError: If the pwm value is not in the range of 0-255.
        """
        if not 0 <= pwm <= 255:
            raise ValueError('PWM value must be in the range of 0-255.')
        self._pwm = pwm
        self._servo = None
        self._can.send(self.id, self._CMD_SET_PWM, 'B', pwm)



# -*-coding:utf-8 -*-
#
# Apache License, Version 2.0
#
# Copyright (c) 2023 Robottime(Beijing) Technology Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Components of robottime devices.

Loads components from entry_points and provides a factory function for
components.
"""

import sys

from robodyno.interfaces import InterfaceType, get_interface_type, WEBOTS_AVAILABLE
from robodyno.components.can_bus.can_bus_device import CanBusDevice

if WEBOTS_AVAILABLE:
    from robodyno.components.webots.webots_device import WebotsDevice

if sys.version_info < (3, 10):
    from importlib_metadata import entry_points
else:
    from importlib.metadata import entry_points

__thismodule = sys.modules[__name__]
__components = {}


def _load_components(eps: list, interface_type: InterfaceType) -> None:
    for ep in eps:
        if ep.name in __components:
            __components[ep.name].update({interface_type: ep.load()})
        else:
            __components.update({ep.name: {interface_type: ep.load()}})


_load_components(
    entry_points(group='robodyno.components.can_bus'), InterfaceType.CAN_BUS
)

if WEBOTS_AVAILABLE:
    _load_components(
        entry_points(group='robodyno.components.webots'), InterfaceType.WEBOTS
    )


def _component_factory(component_name: str) -> callable:
    def construct_component(interface: InterfaceType, *args, **kwargs) -> callable:
        interface_type = get_interface_type(interface)
        if component_name in __components:
            if interface_type in __components[component_name]:
                return __components[component_name][interface_type](
                    interface, *args, **kwargs
                )
            else:
                raise ValueError(
                    f'Component {component_name} is not available for interface '
                    f'type {interface_type}.'
                )
        else:
            raise ValueError(f'Component {component_name} is not available.')

    return construct_component


for name in __components:
    setattr(__thismodule, name, _component_factory(name))



# -*-coding:utf-8 -*-
#
# Apache License, Version 2.0
#
# Copyright (c) 2023 Robottime(Beijing) Technology Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Battery driver.

The battery driver can be used to get the battery voltage and current.

Examples:

    >>> from robodyno.components import Battery
    >>> from robodyno.interfaces import CanBus
    >>> can = CanBus()
    >>> battery = Battery(can)
    >>> print(battery)
    Battery(id=0x0F, voltage=27.66, current=0.0, error=0, rsoc=0.88, fet=3)
"""

from struct import unpack
from typing import Optional
from datetime import datetime

from robodyno.interfaces import CanBus
from robodyno.components import CanBusDevice
from robodyno.components.config.model import Model


class Battery(CanBusDevice):
    """Battery driver.

    Attributes:
        id (int): Device id.
        type (Model): Device type.
        fw_ver (float): Firmware version.
        voltage (float): Battery voltage.
        current (float): Battery current.
        error (int): Error code.
        rsoc (float): Relative state of charge.
        fet (int): FET status.
    """

    _CMD_HEARTBEAT = 0x02
    _CMD_CONFIG_CAN = 0x03
    _CMD_GET_VOLTAGE = 0x04
    _CMD_GET_CURRENT = 0x05
    _CMD_GET_CAPACITY = 0x06
    _CMD_GET_CYTLE_COUNT = 0x07
    _CMD_GET_PRODUCTION_DATE = 0x08
    _CMD_GET_BALANCE_STATUS = 0x09
    _CMD_GET_PROTECTION_STATUS = 0x0A
    _CMD_GET_HARDWARE_VERSION = 0x0B
    _CMD_GET_RSOC = 0x0C
    _CMD_GET_FET = 0x0D
    _CMD_GET_STRING_COUNT = 0x0E
    _CMD_GET_NTC = 0x0F

    def __init__(self, can: CanBus, id_: int = 0x0F):
        """Initializes the battery driver.

        Args:
            can (CanBus): Can bus instance.
            id_ (int): Device id.

        Raises:
            ValueError: If the device id is invalid.
        """
        super().__init__(can, id_)
        self.get_version(timeout=0.015)
        if self.type is None or self.type != Model.ROBODYNO_BATTERY:
            raise ValueError('The device is not a battery supported by robodyno.')
        self.voltage = 0.0
        self.current = 0.0
        self.error = 0
        self.rsoc = 0.0
        self.fet = 0

        self._can.subscribe(
            self._on_hearbeat,
            device_id=self.id,
            cmd_id=self._CMD_HEARTBEAT,
        )

    def __repr__(self) -> str:
        """Gets the string representation of the battery driver.

        Returns:
            str: String representation of the battery driver.
        """
        return (
            f'Battery(id=0x{self.id:02X}, voltage={self.voltage}, '
            f'current={self.current}, error={self.error}, '
            f'rsoc={self.rsoc}, fet={self.fet})'
        )

    def _on_hearbeat(self, data, timestamp, device_id, command_id) -> None:
        """Handles the heartbeat message."""
        del timestamp, device_id, command_id
        voltage, current, error, rsoc, fet = unpack('<HhHBB', data)
        self.voltage = voltage / 100.0
        self.current = current / 100.0
        self.error = error
        self.rsoc = rsoc / 100.0
        self.fet = fet

    def config_can_bus(
        self, new_id: int, heartbeat: int = 1000, bitrate=1000000
    ) -> None:
        """Configures the CAN bus settings.

        Args:
            new_id (int): The new device id.
            heartbeat (int): The heartbeat period in milliseconds.
            bitrate (int): The bitrate of the CAN bus. Choose from 250000,
                500000, 1000000.

        Raises:
            ValueError: If the new CAN id is not in the range of 0x01-0x3f.
        """
        if not 0x01 <= new_id <= 0x3F:
            raise ValueError('New CAN id must be in the range of 0x01-0x3f.')
        bitrate_id = {
            250000: 0,
            500000: 1,
            1000000: 2,
        }.get(bitrate, 2)
        self._can.send(
            self.id, self._CMD_CONFIG_CAN, 'HHI', new_id, bitrate_id, int(heartbeat)
        )

    def get_voltage(self, timeout: Optional[float] = None) -> float:
        """Reads the voltage of the battery.

        Args:
            timeout (float): The timeout in seconds.

        Returns:
            (float | None): the voltage(V). Returns None if times out.
        """
        try:
            (voltage,) = self._can.get(self.id, self._CMD_GET_VOLTAGE, 'H', timeout)
        except TimeoutError:
            return None
        return voltage / 100.0

    def get_current(self, timeout: Optional[float] = None) -> float:
        """Reads the current of the battery.

        Args:
            timeout (float): The timeout in seconds.

        Returns:
            (float | None): the current(A). Positive value means charging.
                Returns None if times out.
        """
        try:
            (current,) = self._can.get(self.id, self._CMD_GET_CURRENT, 'h', timeout)
        except TimeoutError:
            return None
        return current / 100.0

    def get_capacity(self, timeout: Optional[float] = None) -> tuple:
        """Reads the capacity of the battery.

        Args:
            timeout (float): The timeout in seconds.

        Returns:
            (tuple | None): (remaining capacity(Ah), standard_capacity).
                Returns None if times out.
        """
        try:
            (remaining, standard) = self._can.get(
                self.id, self._CMD_GET_CAPACITY, 'HH', timeout
            )
        except TimeoutError:
            return None
        return (remaining / 100.0, standard / 100.0)

    def get_cycle_count(self, timeout: Optional[float] = None) -> int:
        """Reads the cycle count of the battery.

        Args:
            timeout (float): The timeout in seconds.

        Returns:
            (int | None): the cycle count. Returns None if times out.
        """
        try:
            (cycle_count,) = self._can.get(
                self.id, self._CMD_GET_CYTLE_COUNT, 'H', timeout
            )
        except TimeoutError:
            return None
        return cycle_count

    def get_production_date(self, timeout: Optional[float] = None) -> datetime:
        """Reads the production date of the battery.

        Args:
            timeout (float): The timeout in seconds.

        Returns:
            (datetime | None): the production date. Returns None if times out.
        """
        try:
            (time_data,) = self._can.get(
                self.id, self._CMD_GET_PRODUCTION_DATE, 'H', timeout
            )
        except TimeoutError:
            return None
        year = time_data >> 9
        month = (time_data >> 5) & 0x0F
        day = time_data & 0x1F
        return datetime(year, month, day)

    def get_balance_status(self, timeout: Optional[float] = None) -> int:
        """Reads the balance status of the battery.

        Args:
            timeout (float): The timeout in seconds.

        Returns:
            (int | None): the balance status. Returns None if times out.
        """
        try:
            (status,) = self._can.get(
                self.id, self._CMD_GET_BALANCE_STATUS, 'I', timeout
            )
        except TimeoutError:
            return None
        return status

    def get_protection_status(self, timeout: Optional[float] = None) -> int:
        """Reads the protection status of the battery.

        Args:
            timeout (float): The timeout in seconds.

        Returns:
            (int | None): the protection status. Returns None if times out.
        """
        try:
            (status,) = self._can.get(
                self.id, self._CMD_GET_PROTECTION_STATUS, 'H', timeout
            )
        except TimeoutError:
            return None
        return status

    def get_hardware_version(self, timeout: Optional[float] = None) -> str:
        """Reads the hardware version of the battery.

        Args:
            timeout (float): The timeout in seconds.

        Returns:
            (str | None): the hardware version. Returns None if times out.
        """
        try:
            (version,) = self._can.get(
                self.id, self._CMD_GET_HARDWARE_VERSION, 'B', timeout
            )
        except TimeoutError:
            return None
        return f'{version >> 4}.{version & 0x0F}'

    def get_rsoc(self, timeout: Optional[float] = None) -> float:
        """Reads the rsoc of the battery.

        Args:
            timeout (float): The timeout in seconds.

        Returns:
            (float | None): the remaining state of charge. Returns None if times out.
        """
        try:
            (rsoc,) = self._can.get(self.id, self._CMD_GET_RSOC, 'B', timeout)
        except TimeoutError:
            return None
        return rsoc / 100.0

    def get_fet(self, timeout: Optional[float] = None) -> int:
        """Reads the fet status of the battery.

        Args:
            timeout (float): The timeout in seconds.

        Returns:
            (int | None): the fet status. Returns None if times out.
        """
        try:
            (fet,) = self._can.get(self.id, self._CMD_GET_FET, 'B', timeout)
        except TimeoutError:
            return None
        return fet

    def get_string_count(self, timeout: Optional[float] = None) -> int:
        """Reads the string count of the battery.

        Args:
            timeout (float): The timeout in seconds.

        Returns:
            (int | None): the string count. Returns None if times out.
        """
        try:
            (string_count,) = self._can.get(
                self.id, self._CMD_GET_STRING_COUNT, 'B', timeout
            )
        except TimeoutError:
            return None
        return string_count

    def get_temperature(self, timeout: Optional[float] = None) -> tuple:
        """Reads the temperature of the battery.

        Args:
            timeout (float): The timeout in seconds.

        Returns:
            (tuple | None): the temperatures. Returns None if times out.
        """
        try:
            (_, temp1, temp2) = self._can.get(
                self.id, self._CMD_GET_NTC, 'HHH', timeout
            )
        except TimeoutError:
            return None
        return ((temp1 - 2731) / 10.0, (temp2 - 2731) / 10.0)



# -*-coding:utf-8 -*-
#
# Apache License, Version 2.0
#
# Copyright (c) 2023 Robottime(Beijing) Technology Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module provides Webots interface for robodyno devices.

Examples:

```python
from robodyno.interfaces import Webots

def step_callback(webots):
    print(webots.time())

webots = Webots(step_callback=step_callback, start=False)
webots.start()

while True:
    # do something

    if webots.sleep(1) == -1:
        break

webots.stop()
```

"""

from threading import Thread, Lock, Event
from typing import Optional, Callable
import time

try:
    from controller import Robot
except ModuleNotFoundError as e:
    raise RuntimeError('Please run this script in Webots.') from e


class Webots(object):
    """Webots interface for robodyno devices.

    This class provides Webots interface for robodyno devices. It supports
    managing the time and running the simulation.

    A thread is created to run the simulation. When the simulation is running,
    the callback function will be called every time step.

    Attributes:
        robot (Robot): The Webots robot instance.
    """

    def __init__(
        self,
        time_step: Optional[int] = None,
        step_callback: Optional[Callable] = None,
        start: bool = True,
    ):
        """Initialize the Webots interface.

        Args:
            time_step (int, optional): The time step of the simulation. If not
                specified, the default time step of the robot will be used.
            step_callback (callable, optional): The callback function that will
                be called every time step. 
                
                The function should accept one argument, which is the webots
                interface instance. 
                
                The function should not block for too long,
                otherwise the simulation will be delayed.

            start (bool, optional): Whether to start the simulation
                automatically. Defaults to True.
        """
        self.robot = Robot()
        if time_step is None:
            self._time_step = int(self.robot.getBasicTimeStep())
        else:
            self._time_step = int(time_step)

        self._step_callback = step_callback
        self._step_lock = Lock()
        self._step_event = Event()
        self._step_thread = None

        self._devices = set()

        if start:
            self.start()

    @property
    def time_step(self) -> int:
        """Get the time step of the simulation.

        Returns:
            int: The time step of the simulation.
        """
        return self._time_step

    @property
    def step_lock(self) -> Lock:
        """Get the step lock of the simulation.

        Returns:
            Lock: The step lock of the simulation.
        """
        return self._step_lock

    def start(self) -> None:
        """Start the simulation."""
        if self._step_thread is not None:
            return

        self._step_thread = Thread(target=self._step_loop)
        self._step_event.set()
        self._step_thread.start()

    def stop(self) -> None:
        """Stop the simulation."""
        if self._step_thread is None:
            return

        self._step_event.clear()
        self._step_thread.join()
        self._step_thread = None

    def time(self) -> float:
        """Get the current time of the simulation.

        Returns:
            float: The current time of the simulation.
        """
        return self.robot.getTime()

    def sleep(self, seconds: float) -> int:
        """Sleep for a specified amount of time.

        Args:
            seconds (float): The amount of time to sleep.

        Returns:
            (int): 0 if the simulation is running, -1 if the simulation is
                stopped.
        """
        start_time = self.time()
        while self._step_event.is_set():
            if self.time() - start_time >= seconds:
                break
        else:
            return -1
        return 0

    def register(self, device) -> None:
        """Register a device to the simulation.

        Args:
            device (WebotsDevice): The device to register.
        """
        self._devices.add(device)

    def deregister(self, device) -> None:
        """Deregister a device from the simulation.

        Args:
            device (WebotsDevice): The device to deregister.
        """
        if device in self._devices:
            self._devices.remove(device)

    def _step_loop(self) -> None:
        """The loop function of the simulation."""
        while self._step_event.is_set():
            self._step_lock.acquire()

            if self.robot.stepBegin(self._time_step) == -1:
                break
            for device in self._devices:
                device.parallel_step()
            if self.robot.stepEnd() == -1:
                break
            for device in self._devices:
                device.step()

            self._step_lock.release()

            if self._step_callback is not None:
                self._step_callback(self)

            time.sleep(0.001)  # sleep for 1ms to avoid busy loop

        if self._step_lock.locked():
            self._step_lock.release()

        self._step_event.clear()



# -*-coding:utf-8 -*-
#
# Apache License, Version 2.0
#
# Copyright (c) 2023 Robottime(Beijing) Technology Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provides a class to control webots motor.

The webots motor has the same interface as the can bus motor. The webots motor
is used for simulation.

Examples:

```python
from robodyno.components import Motor
from robodyno.interfaces import Webots
webots = Webots()
motor = Motor(webots, 0x10)

motor.position_track_mode(5, 2, 2)
motor.enable()
motor.set_pos(31.4)

while True:
    print(f'pos: {motor.get_pos():.2f}, vel: {motor.get_vel():.2f}')
    if webots.sleep(0.5) == -1:
        break
```
"""

from math import pi, fabs, copysign, sqrt
from enum import Enum
from typing import Optional

from robodyno.components import WebotsDevice
from robodyno.components.can_bus.motor import Motor as CanMotor
from robodyno.components.config.model import Model
from robodyno.components.config.robottime_config import ROBOTTIME_PARAMS
from robodyno.interfaces import Webots, CanBus


class _MotorPositionTrackController(object):
    """Motor position track controller."""

    def __init__(self, max_vel, max_acc, max_dec):
        self.max_vel = max_vel
        self.max_acc = max_acc
        self.max_dec = max_dec
        self.target = 0
        self.pos = 0
        self.vel = 0
        self.top_vel = 0
        self.acc = 0
        self.dec = 0
        self.t = 0
        self.t_acc = 0
        self.t_dec = 0
        self.t_vel = 0

    def set_pos(self, target_pos, pos, vel):
        self.t = 0
        self.target = target_pos
        self.pos = pos
        self.vel = vel

        dx = self.target - self.pos
        self.top_vel = self.max_vel if dx >= 0 else -self.max_vel
        dx_stop = self.vel * fabs(self.vel) / (2.0 * self.max_dec)
        s = copysign(1, dx - dx_stop)
        self.top_vel = s * self.max_vel
        self.acc = s * self.max_acc
        self.dec = -s * self.max_dec
        if s * self.vel > s * self.top_vel:
            self.acc = -self.acc
        self.t_acc = (self.top_vel - self.vel) / self.acc
        self.t_dec = -self.top_vel / self.dec
        dx_min = (
            0.5 * self.t_acc * (self.top_vel + self.vel)
            + 0.5 * self.t_dec * self.top_vel
        )
        if s * dx < s * dx_min:
            self.top_vel = s * sqrt(
                max(
                    (self.dec * self.vel * self.vel + 2 * self.acc * self.dec * dx)
                    / (self.dec - self.acc),
                    0,
                )
            )
            self.t_acc = max(0, (self.top_vel - self.vel) / self.acc)
            self.t_dec = max(0, -self.top_vel / self.dec)
            self.t_vel = 0
        else:
            self.t_vel = (dx - dx_min) / self.top_vel

    def update(self, dt):
        self.t += dt
        if self.t < 0:
            return (self.pos, self.vel)
        if self.t < self.t_acc:
            return (
                self.pos + self.vel * self.t + 0.5 * self.acc * self.t * self.t,
                self.vel + self.acc * self.t,
            )
        if self.t < self.t_acc + self.t_vel:
            return (
                self.pos
                + self.vel * self.t_acc
                + 0.5 * self.acc * self.t_acc * self.t_acc
                + self.top_vel * (self.t - self.t_acc),
                self.top_vel,
            )
        if self.t < self.t_acc + self.t_vel + self.t_dec:
            td = self.t - (self.t_acc + self.t_vel + self.t_dec)
            return (self.target + 0.5 * self.dec * td * td, self.dec * td)
        else:
            return (self.target, self.max_vel)


class _MotorPositionFilterController(object):
    """Motor position filter controller."""

    def __init__(self, bandwidth):
        self.bandwidth = bandwidth
        self.target = 0

    def set_pos(self, target_pos):
        self.target = target_pos

    def update(self, dt, pos, vel):
        bandwidth = min(self.bandwidth, 0.25 / dt)
        input_filter_ki = 2.0 * bandwidth
        input_filter_kp = 0.25 * (input_filter_ki * input_filter_ki)
        delta_pos = self.target - pos
        delta_vel = -vel
        accel = input_filter_kp * delta_pos + input_filter_ki * delta_vel
        new_vel = vel + dt * accel
        new_pos = pos + dt * new_vel
        return (new_pos, new_vel)


class Motor(WebotsDevice):
    """Webots motor component.

    Attributes:
        id (int): ID of the motor.
        type (Model): Type of the motor.
        reduction (float): Reduction of the motor.
        available_velocity (float): The maximum velocity of the motor.
        available_torque (float): The maximum torque of the motor.
        available_current (float): The maximum phase current of the motor.
        torque_constant (float):
            The torque constant of the motor in Nm/A. This is the torque provided by
            the motor when the current is 1A.
        with_brake (bool): Whether the motor has a brake.
        state (Motor.MotorState): The working state of the motor.
        mode (Motor.MotorControlMode): The control mode of the motor.
    """

    def __init__(
        self,
        webots: Webots,
        id_: int = 0x10,
        type_: Optional[str] = None,
        twin: Optional[CanBus] = None,
    ):
        """Initialize the motor.

        Args:
            webots (Webots): The instance of Webots class.
            id_ (int, optional): ID of the motor. Defaults to 0x10.
            type_ (Optional[str], optional): Force the motor type. Defaults to None.
            twin (Optional[CanBus], optional): The twin motor. Defaults to None.

        Raises:
            ValueError: The motor type is invalid.
        """
        webots.step_lock.acquire()
        super().__init__(webots, id_)
        try:
            self._name = f'0x{id_:02X}::motor'
        except TypeError:
            self._name = id_
        self._motor = self._webots.robot.getDevice(self._name)
        self._pos_sensor = self._motor.getPositionSensor()
        self._twin_motor = None
        if twin and isinstance(twin, CanBus):
            self._twin_motor = CanMotor(twin, id_, type_)
            self._init_twin_offset()
            self.type = self._twin_motor.type
        elif type_:
            self.type = getattr(Model, type_.upper(), None)
        else:
            self.type = getattr(Model, self._pos_sensor.getName(), None)
        if self.type not in ROBOTTIME_PARAMS['motor']:
            raise ValueError('Motor type is invalid')
        self.mode = self.MotorControlMode.UNKNOWN
        self.__dict__.update(ROBOTTIME_PARAMS['motor'].get(self.type, None))
        self.state = self.MotorState.DISABLED
        self._rot_factor = self.reduction / 2.0 / pi

        self._motor.enableTorqueFeedback(int(self._webots.time_step))
        self._pos_sensor.enable(int(self._webots.time_step))

        self._track_controller = _MotorPositionTrackController(
            fabs(4 * 2 * pi / self.reduction),
            fabs(10 * 2 * pi / self.reduction),
            fabs(10 * 2 * pi / self.reduction),
        )
        self._filter_controller = _MotorPositionFilterController(10)

        self._input_pos = 0
        self._input_vel = 0
        self._input_torque = 0
        self._max_vel = self.available_velocity
        self._pos_feedback = 0
        self._vel_feedback = 0
        self._torque_feedback = 0
        self._prev_pos_feedback = 0
        self._prev_update_time = self._webots.time()
        self._webots.register(self)
        webots.step_lock.release()
        if self._twin_motor:
            self.twin_mode()
        else:
            self.position_mode()

    def _init_twin_offset(self) -> bool:
        """Initialize the twin motor offset.

        Returns:
            bool: Whether the twin motor offset is initialized successfully.
        """
        self._twin_offset = 0
        if self._twin_motor:
            if self._twin_motor.fw_ver >= 1:
                abs_pos = self._twin_motor.get_abs_pos(0.5)
                if abs_pos:
                    self._twin_offset = abs_pos
                    return True
                else:
                    return False
            else:
                pos = self._twin_motor.get_pos(0.5)
                if pos:
                    self._twin_offset = pos
                    return True
                else:
                    return False

    def _set_velocity(self, velocity: float) -> None:
        """Set the velocity of the motor.

        Args:
            velocity (float): The velocity of the motor.
        """
        self._motor.setVelocity(min(velocity, self._max_vel))

    def parallel_step(self) -> None:
        """See base class."""
        if self.mode == self.MotorControlMode.TWIN_MODE:
            feedback = self._twin_motor.get_feedback(0.1)
            if feedback is not None:
                self._pos_feedback, self._vel_feedback, self._torque_feedback = feedback
                self._pos_feedback -= self._twin_offset
            if self._twin_motor.fw_ver >= 1:
                abs_pos = self._twin_motor.get_abs_pos(0.1)
                if abs_pos:
                    self._pos_feedback = abs_pos - self._twin_offset

    def step(self) -> None:
        """See base class."""
        if self.mode == self.MotorControlMode.TWIN_MODE:
            self._motor.setPosition(self._pos_feedback)
            return

        t = self._webots.time()
        dt = t - self._prev_update_time
        self._prev_update_time = t
        if dt == 0:
            return
        self._pos_feedback = self._pos_sensor.getValue()
        self._vel_feedback = (self._pos_feedback - self._prev_pos_feedback) / dt
        self._prev_pos_feedback = self._pos_feedback
        self._torque_feedback = self._motor.getTorqueFeedback()
        if self.state == self.MotorState.DISABLED:
            self._motor.setTorque(0)
        elif self.state == self.MotorState.ENABLED:
            if self.mode == self.MotorControlMode.TORQUE_MODE:
                self._motor.setTorque(self._input_torque)
            elif (
                self.mode == self.MotorControlMode.VELOCITY_MODE
                or self.mode == self.MotorControlMode.VELOCITY_RAMP_MODE
            ):
                self._set_velocity(self._input_vel)
            elif self.mode == self.MotorControlMode.POSITION_MODE:
                self._motor.setPosition(self._input_pos)
            elif self.mode == self.MotorControlMode.POSITION_FILTER_MODE:
                p, v = self._filter_controller.update(
                    dt, self._pos_feedback, self._vel_feedback
                )
                self._motor.setPosition(p)
                self._set_velocity(fabs(v))
            elif self.mode == self.MotorControlMode.POSITION_TRACK_MODE:
                p, v = self._track_controller.update(dt)
                self._motor.setPosition(p)
                self._set_velocity(fabs(v))

    def enable(self) -> None:
        """Enables the motor."""
        self._webots.step_lock.acquire()
        self.state = self.MotorState.ENABLED
        self._webots.step_lock.release()

    def disable(self) -> None:
        """Disables the motor."""
        self._webots.step_lock.acquire()
        self.state = self.MotorState.DISABLED
        self._webots.step_lock.release()

    def twin_mode(self) -> None:
        """Sets the motor to digital twin mode."""
        self._webots.step_lock.acquire()
        self._motor.setPosition(self._pos_feedback)
        self._set_velocity(self._max_vel)
        self._motor.setAcceleration(-1)
        self.mode = self.MotorControlMode.TWIN_MODE
        self.state = self.MotorState.ENABLED
        self._webots.step_lock.release()

    def position_mode(self) -> None:
        """Sets the motor to position mode."""
        self._webots.step_lock.acquire()
        self._motor.setPosition(self._pos_feedback)
        self._set_velocity(self._max_vel)
        self._motor.setAcceleration(-1)
        self.mode = self.MotorControlMode.POSITION_MODE
        self._webots.step_lock.release()

    def position_filter_mode(self, bandwidth: float) -> None:
        """Sets the motor to position filter mode.

        Args:
            bandwidth (float): The bandwidth of the filter in Hz. The value is suggested
                to be the control loop frequency in Hz.
        """
        self._webots.step_lock.acquire()
        self._filter_controller = _MotorPositionFilterController(bandwidth)
        self._filter_controller.set_pos(self._pos_feedback)
        self._motor.setAcceleration(-1)
        self.mode = self.MotorControlMode.POSITION_FILTER_MODE
        self._webots.step_lock.release()

    def position_track_mode(self, vel: float, acc: float, dec: float) -> None:
        """Sets the motor to position track mode.

        Args:
            vel (float): The velocity limit in rad/s.
            acc (float): The acceleration in rad/s^2. The value must be positive.
            dec (float): The deceleration in rad/s^2. The value must be positive.
        """
        self._webots.step_lock.acquire()
        self._track_controller = _MotorPositionTrackController(vel, acc, dec)
        self._track_controller.set_pos(
            self._pos_feedback, self._pos_feedback, self._vel_feedback
        )
        self._motor.setAcceleration(-1)
        self.mode = self.MotorControlMode.POSITION_TRACK_MODE
        self._webots.step_lock.release()

    def velocity_mode(self) -> None:
        """Sets the motor to velocity mode."""
        self._webots.step_lock.acquire()
        self._motor.setPosition(float('+inf'))
        self._set_velocity(self._input_vel)
        self._motor.setAcceleration(-1)
        self.mode = self.MotorControlMode.VELOCITY_MODE
        self._webots.step_lock.release()

    def velocity_ramp_mode(self, ramp: float) -> None:
        """Sets the motor to velocity ramp mode.

        Args:
            ramp (float): The velocity ramp in rad/s^2. The value must be positive.
        """
        self._webots.step_lock.acquire()
        self._motor.setPosition(float('+inf'))
        self._set_velocity(self._input_vel)
        self._motor.setAcceleration(ramp)
        self.mode = self.MotorControlMode.VELOCITY_RAMP_MODE
        self._webots.step_lock.release()

    def torque_mode(self) -> None:
        """Sets the motor to torque mode."""
        self._webots.step_lock.acquire()
        self._motor.setTorque(self._input_torque)
        self.mode = self.MotorControlMode.TORQUE_MODE
        self._webots.step_lock.release()

    def set_pos(self, pos: float, vel_ff: float = 0, torque_ff: float = 0) -> None:
        """Sets the target position of the motor.

        velocity feedforward and torque feedforward are not supported in Webots.

        Args:
            pos (float): The target position in rad.
        """
        del vel_ff, torque_ff
        self._webots.step_lock.acquire()
        self._input_pos = pos
        self._filter_controller.set_pos(pos)
        self._track_controller.set_pos(pos, self._pos_feedback, self._vel_feedback)
        self._webots.step_lock.release()

    def set_abs_pos(self, pos: float, vel_ff: float = 0, torque_ff: float = 0) -> None:
        """Sets the target absolute position of the motor.

        velocity feedforward and torque feedforward are not supported in Webots.

        Args:
            pos (float): The target absolute position in rad.
        """
        del vel_ff, torque_ff
        self._webots.step_lock.acquire()
        self._input_pos = pos
        self._filter_controller.set_pos(pos)
        self._track_controller.set_pos(pos, self._pos_feedback, self._vel_feedback)
        self._webots.step_lock.release()

    def set_vel(self, vel: float, torque_ff: float = 0) -> None:
        """Sets the target velocity of the motor.

        Args:
            vel (float): The target velocity in rad/s.
        """
        del torque_ff
        self._webots.step_lock.acquire()
        self._input_vel = vel
        self._webots.step_lock.release()

    def set_torque(self, torque: float) -> None:
        """Sets the target torque of the motor.

        Args:
            torque (float): The target torque in Nm.
        """
        self._webots.step_lock.acquire()
        self._input_torque = torque
        self._webots.step_lock.release()

    def set_vel_limit(self, vel_lim: float) -> None:
        """Sets the velocity limit of the motor.

        Args:
            vel_lim (float): The velocity limit in rad/s. The value must be positive.

        Raises:
            ValueError: If the velocity limit is out of range.
        """
        self._webots.step_lock.acquire()
        if not 0 < fabs(vel_lim) <= self.available_velocity:
            raise ValueError(f'Velocity limit {vel_lim} is out of range.')
        self._max_vel = fabs(vel_lim)
        if self.mode in [
            self.MotorControlMode.POSITION_MODE,
            self.MotorControlMode.POSITION_FILTER_MODE,
            self.MotorControlMode.POSITION_TRACK_MODE,
        ]:
            self._set_velocity(fabs(vel_lim))
        self._webots.step_lock.release()

    def set_current_limit(self, current_lim: float) -> None:
        """Sets the current limit of the motor.

        Args:
            current_lim (float): The current limit in amps. The value must be positive.

        Raises:
            ValueError: If the current limit is out of range.
        """
        self._webots.step_lock.acquire()
        if not 0 < fabs(current_lim) <= self.available_current:
            raise ValueError(f'Current limit {current_lim} is out of range.')
        self._motor.setAvailableTorque(fabs(current_lim) * self.torque_constant)
        self._webots.step_lock.release()

    def get_mode(self, timeout: Optional[float] = None) -> Optional[tuple]:
        """Reads the control mode and the mode parameters of the motor.

        Timeout is not supported in Webots.

        Returns:
            (tuple): The control mode and the mode parameters.
        """
        del timeout
        ret = None
        if self.mode == self.MotorControlMode.POSITION_FILTER_MODE:
            ret = (self.mode, {'bandwidth': self._filter_controller.bandwidth})
        elif self.mode == self.MotorControlMode.POSITION_TRACK_MODE:
            ret = (
                self.mode,
                {
                    'vel': fabs(self._track_controller.max_vel / self._rot_factor),
                    'acc': fabs(self._track_controller.max_acc / self._rot_factor),
                    'dec': fabs(self._track_controller.max_dec / self._rot_factor),
                },
            )
        elif self.mode == self.MotorControlMode.VELOCITY_RAMP_MODE:
            ret = (self.mode, {'ramp': self._motor.getAcceleration()})
        else:
            ret = (self.mode,)
        return ret

    def get_pos(self, timeout: Optional[float] = None) -> float:
        """Reads the position of the motor.

        Timeout is not supported in Webots.

        Returns:
            (float): The position in rad.
        """
        del timeout
        return self._pos_feedback

    def get_abs_pos(self, timeout: Optional[float] = None) -> float:
        """Reads the absolute position of the motor.

        Timeout is not supported in Webots.

        Returns:
            (float): The absolute position in rad.
        """
        del timeout
        return self._pos_feedback

    def get_vel(self, timeout: Optional[float] = None) -> float:
        """Reads the velocity of the motor.

        Timeout is not supported in Webots.

        Returns:
            (float): The velocity in rad/s.
        """
        del timeout
        return self._vel_feedback

    def get_torque(self, timeout: Optional[float] = None) -> float:
        """Reads the torque of the motor.

        Timeout is not supported in Webots.

        Returns:
            (float): The torque in Nm.
        """
        del timeout
        return self._torque_feedback

    def get_vel_limit(self, timeout: Optional[float] = None) -> float:
        """Reads the velocity limit of the motor.

        Timeout is not supported in Webots.

        Returns:
            (float): The velocity limit in rad/s.
        """
        del timeout
        return self._max_vel

    def get_current_limit(self, timeout: Optional[float] = None) -> float:
        """Reads the current limit of the motor.

        Timeout is not supported in Webots.

        Returns:
            (float): The current limit in amps.
        """
        del timeout
        return self._motor.getAvailableTorque() / self.torque_constant

    class MotorState(Enum):
        """Enumerates motor working states."""

        UNKNOWN = -1
        """The state of the motor is unknown."""
        DISABLED = 1
        """The motor is disabled."""
        CALIBRATE = 3
        """The motor is preparing for calibration."""
        MOTOR_CALIBRATING = 4
        """The motor is calibrating."""
        OFFSET_CALIBRATING = 7
        """The motor is calibrating the encoder."""
        ENABLED = 8
        """The motor is enabled."""
        HOMING = 11
        """The motor is aligning the encoders."""

    class MotorControlMode(Enum):
        """Enumerates motor control modes."""

        UNKNOWN = (0, 0)
        """The motor control mode is unknown."""
        POSITION_MODE = (3, 1)
        """Control the motor in position mode."""
        POSITION_FILTER_MODE = (3, 3)
        """Control the motor in position filter mode."""
        POSITION_TRACK_MODE = (3, 5)
        """Control the motor in position track mode."""
        VELOCITY_MODE = (2, 1)
        """Control the motor in velocity mode."""
        VELOCITY_RAMP_MODE = (2, 2)
        """Control the motor in velocity ramp mode."""
        TORQUE_MODE = (1, 1)
        """Control the motor in torque mode."""
        TWIN_MODE = (255, 255)
        """Digital twin mode."""
